@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2022 Derick Rethans                               |
+   | Copyright (c) 2002-2023 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -36,6 +36,9 @@
 
 #include "base.h"
 #include "filter.h"
+#ifdef __linux__
+# include "ctrl_socket.h"
+#endif
 #include "develop/develop.h"
 #include "develop/stack.h"
 #include "gcstats/gc_stats.h"
@@ -98,7 +101,7 @@ static zend_op_array *xdebug_compile_file(zend_file_handle *file_handle, int typ
 void xdebug_func_dtor_by_ref(xdebug_func *elem)
 {
 	if (elem->function) {
-		xdfree(elem->function);
+		zend_string_release(elem->function);
 	}
 	if (elem->object_class) {
 		zend_string_release(elem->object_class);
@@ -106,6 +109,9 @@ void xdebug_func_dtor_by_ref(xdebug_func *elem)
 	if (elem->scope_class) {
 		zend_string_release(elem->scope_class);
 	}
+	if (elem->include_filename) {
+		zend_string_release(elem->include_filename);
+	}
 }
 
 void xdebug_func_dtor(xdebug_func *elem)
@@ -135,10 +141,6 @@ static void function_stack_entry_dtor(void *elem)
 		xdfree(e->var);
 	}
 
-	if (e->include_filename) {
-		zend_string_release(e->include_filename);
-	}
-
 	if (e->declared_vars) {
 		xdebug_llist_destroy(e->declared_vars, NULL);
 		e->declared_vars = NULL;
@@ -210,7 +212,7 @@ void xdebug_build_fname(xdebug_func *tmp, zend_execute_data *edata)
 
 	if (edata && edata->func && edata->func == (zend_function*) &zend_pass_function) {
 		tmp->type     = XFUNC_ZEND_PASS;
-		tmp->function = xdstrdup("{zend_pass}");
+		tmp->function = ZSTR_INIT_LITERAL("{zend_pass}", false);
 	} else if (edata && edata->func) {
 		tmp->type = XFUNC_NORMAL;
 		if ((Z_TYPE(edata->This)) == IS_OBJECT) {
@@ -238,8 +240,8 @@ void xdebug_build_fname(xdebug_func *tmp, zend_execute_data *edata)
 		}
 		if (edata->func->common.function_name) {
 			if (edata->func->common.fn_flags & ZEND_ACC_CLOSURE) {
-				tmp->function = xdebug_wrap_closure_location_around_function_name(&edata->func->op_array, STR_NAME_VAL(edata->func->common.function_name));
-			} else if (strncmp(edata->func->common.function_name->val, "call_user_func", 14) == 0) {
+				tmp->function = xdebug_wrap_closure_location_around_function_name(&edata->func->op_array, edata->func->common.function_name);
+			} else if (strncmp(ZSTR_VAL(edata->func->common.function_name), "call_user_func", 14) == 0) {
 				zend_string *fname = NULL;
 				int          lineno = 0;
 
@@ -262,15 +264,16 @@ void xdebug_build_fname(xdebug_func *tmp, zend_execute_data *edata)
 
 				lineno = find_line_number_for_current_execute_point(edata);
 
-				tmp->function = xdebug_sprintf(
+				tmp->function = zend_strpprintf(
+					0,
 					"%s:{%s:%d}",
-					edata->func->common.function_name->val,
+					ZSTR_VAL(edata->func->common.function_name),
 					ZSTR_VAL(fname),
 					lineno
 				);
 			} else {
 normal_after_all:
-				tmp->function = xdstrdup(edata->func->common.function_name->val);
+				tmp->function = zend_string_copy(edata->func->common.function_name);
 			}
 		} else if (
 			edata &&
@@ -285,7 +288,7 @@ normal_after_all:
 			)
 		) {
 			tmp->type = XFUNC_NORMAL;
-			tmp->function = xdstrdup("{internal eval}");
+			tmp->function = ZSTR_INIT_LITERAL("{internal eval}", false);
 		} else if (
 			edata &&
 			edata->prev_execute_data &&
@@ -665,7 +668,7 @@ function_stack_entry *xdebug_add_stack_frame(zend_execute_data *zdata, zend_op_a
 
 	xdebug_build_fname(&(tmp->function), zdata);
 	if (!tmp->function.type) {
-		tmp->function.function     = xdstrdup("{main}");
+		tmp->function.function     = ZSTR_INIT_LITERAL("{main}", false);
 		tmp->function.object_class = NULL;
 		tmp->function.scope_class  = NULL;
 		tmp->function.type         = XFUNC_MAIN;
@@ -680,9 +683,9 @@ function_stack_entry *xdebug_add_stack_frame(zend_execute_data *zdata, zend_op_a
 		}
 
 		if (tmp->function.type == XFUNC_EVAL && XG_BASE(last_eval_statement)) {
-			tmp->include_filename = zend_string_copy(XG_BASE(last_eval_statement));
+			tmp->function.include_filename = zend_string_copy(XG_BASE(last_eval_statement));
 		} else {
-			tmp->include_filename = zend_string_copy(zend_get_executed_filename_ex());
+			tmp->function.include_filename = zend_string_copy(zend_get_executed_filename_ex());
 		}
 	} else {
 		tmp->lineno = find_line_number_for_current_execute_point(edata);
@@ -707,15 +710,14 @@ function_stack_entry *xdebug_add_stack_frame(zend_execute_data *zdata, zend_op_a
 	return tmp;
 }
 
-static void xdebug_execute_ex(zend_execute_data *execute_data)
+/** Function interceptors and dispatchers to modules ***********************/
+
+static void xdebug_execute_user_code_begin(zend_execute_data *execute_data)
 {
-	zend_op_array        *op_array = &(execute_data->func->op_array);
-	zend_execute_data    *edata = execute_data->prev_execute_data;
+	zend_op_array     *op_array = &(execute_data->func->op_array);
+	zend_execute_data *edata = execute_data->prev_execute_data;
+
 	function_stack_entry *fse;
-	int                   function_nr = 0;
-	char                 *code_coverage_function_name = NULL;
-	zend_string          *code_coverage_filename = NULL;
-	int                   code_coverage_init = 0;
 
 	/* For PHP 7, we need to reset the opline to the start, so that all opcode
 	 * handlers are being hit. But not for generators, as that would make an
@@ -724,29 +726,6 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 		EX(opline) = EX(func)->op_array.opcodes;
 	}
 
-	if (xdebug_debugger_bailout_if_no_exec_requested()) {
-		return;
-	}
-
-	/* If we're evaluating for the debugger's eval capability, just bail out */
-	if (op_array && op_array->filename && strcmp("xdebug://debug-eval", STR_NAME_VAL(op_array->filename)) == 0) {
-		xdebug_old_execute_ex(execute_data);
-		return;
-	}
-
-	/* if we're in a ZEND_EXT_STMT, we ignore this function call as it's likely
-	   that it's just being called to check for breakpoints with conditions */
-	if (edata && edata->func && ZEND_USER_CODE(edata->func->type) && edata->opline && edata->opline->opcode == ZEND_EXT_STMT) {
-		xdebug_old_execute_ex(execute_data);
-		return;
-	}
-
-	/* If the stack vector hasn't been initialised yet, we should abort immediately */
-	if (!XG_BASE(stack)) {
-		xdebug_old_execute_ex(execute_data);
-		return;
-	}
-
 	if (XG_BASE(in_execution) && XDEBUG_VECTOR_COUNT(XG_BASE(stack)) == 0) {
 		if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
 			xdebug_debugger_set_program_name(op_array->filename);
@@ -774,17 +753,19 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 	fse->function.internal = 0;
 
 	/* A hack to make __call work with profiles. The function *is* user defined after all. */
-	if (fse && xdebug_vector_element_is_valid(XG_BASE(stack), fse - 1) && fse->function.function && (strcmp(fse->function.function, "__call") == 0)) {
+	if (fse && xdebug_vector_element_is_valid(XG_BASE(stack), fse - 1) && fse->function.function && zend_string_equals_literal(fse->function.function, "__call")) {
 		(fse - 1)->user_defined = XDEBUG_USER_DEFINED;
 	}
 
-	function_nr = XG_BASE(function_count);
+#ifdef __linux__
+	xdebug_control_socket_dispatch();
+#endif
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
 		xdebug_monitor_handler(fse);
 	}
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_TRACING)) {
-		xdebug_tracing_execute_ex(function_nr, fse);
+		xdebug_tracing_execute_ex(fse);
 	}
 
 	fse->execute_data = EG(current_execute_data)->prev_execute_data;
@@ -792,26 +773,8 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 		fse->symbol_table = EG(current_execute_data)->symbol_table;
 	}
 
-	if (XG_BASE(stack) && (XINI_DEV(show_local_vars) || xdebug_is_debug_connection_active())) {
-		/* Because include/require is treated as a stack level, we have to add used
-		 * variables in include/required files to all the stack levels above, until
-		 * we hit a function or the top level stack.  This is so that the variables
-		 * show up correctly where they should be.  We always call
-		 * xdebug_lib_register_compiled_variables on the current stack level,
-		 * otherwise vars in include files do not show up in the locals list. */
-		function_stack_entry *loop_fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
-		int                   i;
-
-		for (i = 0; i < XDEBUG_VECTOR_COUNT(XG_BASE(stack)); i++, loop_fse--) {
-			xdebug_lib_register_compiled_variables(loop_fse, op_array);
-			if (XDEBUG_IS_NORMAL_FUNCTION(&loop_fse->function)) {
-				break;
-			}
-		}
-	}
-
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_COVERAGE)) {
-		code_coverage_init = xdebug_coverage_execute_ex(fse, op_array, &code_coverage_filename, &code_coverage_function_name);
+		fse->code_coverage_init = xdebug_coverage_execute_ex(fse, op_array, &fse->code_coverage_filename, &fse->code_coverage_function_name);
 	}
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
@@ -827,29 +790,31 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_PROFILING)) {
 		xdebug_profiler_execute_ex(fse, op_array);
 	}
+}
 
-	xdebug_old_execute_ex(execute_data);
+static void xdebug_execute_user_code_end(zend_execute_data *execute_data, zval *retval)
+{
+	zend_op_array        *op_array = &(execute_data->func->op_array);
+	function_stack_entry *fse;
 
-	/* Re-acquire the tail as nested calls through xdebug_old_execute_ex()
-	 * might have reallocated the vector */
 	fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_PROFILING)) {
 		xdebug_profiler_execute_ex_end(fse);
 	}
 
-	if (code_coverage_init) {
-		xdebug_coverage_execute_ex_end(fse, op_array, code_coverage_filename, code_coverage_function_name);
+	if (fse->code_coverage_init) {
+		xdebug_coverage_execute_ex_end(fse, op_array, fse->code_coverage_filename, fse->code_coverage_function_name);
 	}
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_TRACING)) {
-		xdebug_tracing_execute_ex_end(function_nr, fse, execute_data);
+		xdebug_tracing_execute_ex_end(fse, execute_data);
 	}
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
 		zval *return_value = NULL;
 
-		if (!fse->is_trampoline && execute_data->return_value && !(op_array->fn_flags & ZEND_ACC_GENERATOR)) {
+		if (!fse->is_trampoline && retval && !(op_array->fn_flags & ZEND_ACC_GENERATOR)) {
 			return_value = execute_data->return_value;
 		}
 
@@ -857,14 +822,71 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 		xdebug_debugger_handle_breakpoints(fse, XDEBUG_BREAKPOINT_TYPE_RETURN|XDEBUG_BREAKPOINT_TYPE_EXTERNAL, return_value);
 	}
 
-	fse->symbol_table = NULL;
-	fse->execute_data = NULL;
-
 	if (XG_BASE(stack)) {
 		xdebug_vector_pop(XG_BASE(stack));
 	}
 }
 
+static bool should_run_user_handler(zend_execute_data *execute_data)
+{
+	zend_op_array     *op_array = &(execute_data->func->op_array);
+	zend_execute_data *prev_edata = execute_data->prev_execute_data;
+
+	if (xdebug_debugger_bailout_if_no_exec_requested()) {
+		return false;
+	}
+
+	if (!ZEND_USER_CODE(op_array->type)) {
+		return false;
+	}
+
+	/* If we're evaluating for the debugger's eval capability, just bail out */
+	if (op_array && ZEND_USER_CODE(op_array->type) && op_array->filename && op_array->filename && strcmp("xdebug://debug-eval", STR_NAME_VAL(op_array->filename)) == 0) {
+		return false;
+	}
+
+	/* if we're in a ZEND_EXT_STMT, we ignore this function call as it's likely
+	   that it's just being called to check for breakpoints with conditions */
+	if (prev_edata && prev_edata->func && ZEND_USER_CODE(prev_edata->func->type) && prev_edata->opline && prev_edata->opline->opcode == ZEND_EXT_STMT) {
+		return false;
+	}
+
+	/* If the stack vector hasn't been initialised yet, we should abort immediately */
+	if (!XG_BASE(stack)) {
+		return false;
+	}
+
+	return true;
+}
+
+/* This is confusing. On PHP 8.1 we flip the logic, as normal user functions
+ * are handled through the Observer API. Once PHP 8.0 support is dropped, the
+ * negation should be **added** to the usage below in xdebug_execute_ex. */
+static bool should_run_user_handler_wrapper(zend_execute_data *execute_data)
+{
+#if PHP_VERSION_ID >= 80100
+	return !should_run_user_handler(execute_data);
+#else
+	return should_run_user_handler(execute_data);
+#endif
+}
+
+/* We still need this to do "include", "require", and "eval" */
+static void xdebug_execute_ex(zend_execute_data *execute_data)
+{
+	bool run_user_handler = should_run_user_handler_wrapper(execute_data);
+
+	if (run_user_handler) {
+		xdebug_execute_user_code_begin(execute_data);
+	}
+
+	xdebug_old_execute_ex(execute_data);
+
+	if (run_user_handler) {
+		xdebug_execute_user_code_end(execute_data, execute_data->return_value);
+	}
+}
+
 static int check_soap_call(function_stack_entry *fse, zend_execute_data *execute_data)
 {
 	if (
@@ -892,28 +914,25 @@ static int check_soap_call(function_stack_entry *fse, zend_execute_data *execute
 	return 0;
 }
 
-static void xdebug_execute_internal(zend_execute_data *current_execute_data, zval *return_value)
+static bool should_run_internal_handler(zend_execute_data *execute_data)
+{
+	/* If the stack vector hasn't been initialised yet, we should abort immediately */
+	if (!XG_BASE(stack)) {
+		return false;
+	}
+
+	if (!execute_data || !execute_data->func || ZEND_USER_CODE(execute_data->func->type)) {
+		return false;
+	}
+
+	return true;
+}
+
+
+static void xdebug_execute_internal_begin(zend_execute_data *current_execute_data)
 {
 	zend_execute_data    *edata = EG(current_execute_data);
 	function_stack_entry *fse;
-	int                   function_nr = 0;
-	int                   function_call_traced = 0;
-	int                   restore_error_handler_situation = 0;
-#if PHP_VERSION_ID >= 80100
-	void                (*tmp_error_cb)(int type, zend_string *error_filename, const uint32_t error_lineno, zend_string *message) = NULL;
-#else
-	void                (*tmp_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, zend_string *message) = NULL;
-#endif
-
-	/* If the stack vector hasn't been initialised yet, we should abort immediately */
-	if (!XG_BASE(stack)) {
-		if (xdebug_old_execute_internal) {
-			xdebug_old_execute_internal(current_execute_data, return_value);
-		} else {
-			execute_internal(current_execute_data, return_value);
-		}
-		return;
-	}
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP) && (signed long) XDEBUG_VECTOR_COUNT(XG_BASE(stack)) >= XINI_BASE(max_nesting_level) && (XINI_BASE(max_nesting_level) != -1)) {
 		zend_throw_exception_ex(zend_ce_error, 0, "Xdebug has detected a possible infinite loop, and aborted your script with a stack depth of '" ZEND_LONG_FMT "' frames", XINI_BASE(max_nesting_level));
@@ -922,13 +941,16 @@ static void xdebug_execute_internal(zend_execute_data *current_execute_data, zva
 	fse = xdebug_add_stack_frame(edata, &edata->func->op_array, XDEBUG_BUILT_IN);
 	fse->function.internal = 1;
 
-	function_nr = XG_BASE(function_count);
-
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
 		xdebug_monitor_handler(fse);
 	}
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_TRACING)) {
-		function_call_traced = xdebug_tracing_execute_internal(function_nr, fse);
+		fse->function_call_traced = xdebug_tracing_execute_internal(fse);
+	}
+
+	fse->execute_data = EG(current_execute_data)->prev_execute_data;
+	if (ZEND_CALL_INFO(EG(current_execute_data)) & ZEND_CALL_HAS_SYMBOL_TABLE) {
+		fse->symbol_table = EG(current_execute_data)->symbol_table;
 	}
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
@@ -938,20 +960,18 @@ static void xdebug_execute_internal(zend_execute_data *current_execute_data, zva
 
 	/* Check for SOAP */
 	if (check_soap_call(fse, current_execute_data)) {
-		restore_error_handler_situation = 1;
-		tmp_error_cb = zend_error_cb;
+		fse->soap_error_cb = zend_error_cb;
 		xdebug_base_use_original_error_cb();
 	}
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_PROFILING)) {
 		xdebug_profiler_execute_internal(fse);
 	}
+}
 
-	if (xdebug_old_execute_internal) {
-		xdebug_old_execute_internal(current_execute_data, return_value);
-	} else {
-		execute_internal(current_execute_data, return_value);
-	}
+static void xdebug_execute_internal_end(zend_execute_data *current_execute_data, zval *return_value)
+{
+	function_stack_entry *fse;
 
 	/* Re-acquire the tail as nested calls through
 	 * xdebug_old_execute_internal() might have reallocated the vector */
@@ -962,15 +982,15 @@ static void xdebug_execute_internal(zend_execute_data *current_execute_data, zva
 	}
 
 	/* Restore SOAP situation if needed */
-	if (restore_error_handler_situation) {
-		zend_error_cb = tmp_error_cb;
+	if (fse->soap_error_cb) {
+		zend_error_cb = fse->soap_error_cb;
 	}
 
 	/* We only call the function_exit handler and return value handler if the
 	 * function call was also traced. Otherwise we end up with return trace
 	 * lines without a corresponding function call line. */
-	if (XDEBUG_MODE_IS(XDEBUG_MODE_TRACING) && function_call_traced) {
-		xdebug_tracing_execute_internal_end(function_nr, fse, return_value);
+	if (XDEBUG_MODE_IS(XDEBUG_MODE_TRACING) && fse->function_call_traced) {
+		xdebug_tracing_execute_internal_end(fse, return_value);
 	}
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
@@ -983,6 +1003,59 @@ static void xdebug_execute_internal(zend_execute_data *current_execute_data, zva
 	}
 }
 
+#if PHP_VERSION_ID < 80200
+static void xdebug_execute_internal(zend_execute_data *current_execute_data, zval *return_value)
+{
+	bool run_internal_handler = should_run_internal_handler(current_execute_data);
+
+	if (run_internal_handler) {
+		xdebug_execute_internal_begin(current_execute_data);
+	}
+
+	if (xdebug_old_execute_internal) {
+		xdebug_old_execute_internal(current_execute_data, return_value);
+	} else {
+		execute_internal(current_execute_data, return_value);
+	}
+
+	if (run_internal_handler) {
+		xdebug_execute_internal_end(current_execute_data, return_value);
+	}
+}
+#endif
+
+#if PHP_VERSION_ID >= 80100
+static void xdebug_execute_begin(zend_execute_data *execute_data)
+{
+	if (should_run_user_handler(execute_data)) {
+		xdebug_execute_user_code_begin(execute_data);
+	}
+#if PHP_VERSION_ID >= 80200
+	if (should_run_internal_handler(execute_data)) {
+		xdebug_execute_internal_begin(execute_data);
+	}
+#endif
+}
+
+static void xdebug_execute_end(zend_execute_data *execute_data, zval *retval)
+{
+	if (should_run_user_handler(execute_data)) {
+		xdebug_execute_user_code_end(execute_data, retval);
+	}
+#if PHP_VERSION_ID >= 80200
+	if (should_run_internal_handler(execute_data)) {
+		xdebug_execute_internal_end(execute_data, retval);
+	}
+#endif
+}
+
+static zend_observer_fcall_handlers xdebug_observer_init(zend_execute_data *execute_data)
+{
+	return (zend_observer_fcall_handlers){xdebug_execute_begin, xdebug_execute_end};
+}
+#endif
+/***************************************************************************/
+
 static void xdebug_base_overloaded_functions_setup(void)
 {
 	zend_function *orig;
@@ -1070,6 +1143,7 @@ static int xdebug_closure_serialize_deny_wrapper(zval *object, unsigned char **b
 }
 
 #if PHP_VERSION_ID >= 80100
+/** Handling fibers ********************************************************/
 static struct xdebug_fiber_entry* xdebug_fiber_entry_ctor(xdebug_vector *stack)
 {
 	struct xdebug_fiber_entry *tmp = xdmalloc(sizeof(struct xdebug_fiber_entry));
@@ -1085,18 +1159,13 @@ static void xdebug_fiber_entry_dtor(struct xdebug_fiber_entry *entry)
 	xdfree(entry);
 }
 
-static xdebug_str *create_key_for_fiber(zend_fiber_context *fiber)
+static zend_string *create_key_for_fiber(zend_fiber_context *fiber)
 {
-	xdebug_str *tmp = xdebug_str_new();
-
-	xdebug_str_add_fmt(tmp, "{fiber:%0" PRIXPTR "}", ((uintptr_t) fiber));
-
-	return tmp;
+	return zend_strpprintf(0, "{fiber:%0" PRIXPTR "}", ((uintptr_t) fiber));
 }
 
 static void add_fiber_main(zend_fiber_context *fiber)
 {
-	xdebug_str           *key = create_key_for_fiber(fiber);
 	function_stack_entry *tmp = (function_stack_entry*) xdebug_vector_push(XG_BASE(stack));
 
 	tmp->level        = XDEBUG_VECTOR_COUNT(XG_BASE(stack));
@@ -1104,7 +1173,7 @@ static void add_fiber_main(zend_fiber_context *fiber)
 	tmp->function.type = XFUNC_FIBER;
 	tmp->function.object_class = NULL;
 	tmp->function.scope_class = NULL;
-	tmp->function.function = xdstrdup(key->d);
+	tmp->function.function = create_key_for_fiber(fiber);
 	tmp->filename = zend_string_copy(zend_get_executed_filename_ex());
 	tmp->lineno = zend_get_executed_lineno();
 
@@ -1113,40 +1182,38 @@ static void add_fiber_main(zend_fiber_context *fiber)
 	XG_BASE(prev_memory) = tmp->memory;
 
 	tmp->nanotime = xdebug_get_nanotime();
-
-	xdebug_str_free(key);
 }
 
 static xdebug_vector* create_stack_for_fiber(zend_fiber_context *fiber)
 {
 	xdebug_vector             *tmp_stack = xdebug_vector_alloc(sizeof(function_stack_entry), function_stack_entry_dtor);
-	xdebug_str                *key       = create_key_for_fiber(fiber);
+	zend_string               *key       = create_key_for_fiber(fiber);
 	struct xdebug_fiber_entry *entry     = xdebug_fiber_entry_ctor(tmp_stack);
 
-	xdebug_hash_add(XG_BASE(fiber_stacks), key->d, key->l, entry);
+	xdebug_hash_add(XG_BASE(fiber_stacks), ZSTR_VAL(key), ZSTR_LEN(key), entry);
 
-	xdebug_str_free(key);
+	zend_string_release(key);
 
 	return tmp_stack;
 }
 
 static void remove_stack_for_fiber(zend_fiber_context *fiber)
 {
-	xdebug_str *key = create_key_for_fiber(fiber);
+	zend_string *key = create_key_for_fiber(fiber);
 
-	xdebug_hash_delete(XG_BASE(fiber_stacks), key->d, key->l);
+	xdebug_hash_delete(XG_BASE(fiber_stacks), ZSTR_VAL(key), ZSTR_LEN(key));
 
-	xdebug_str_free(key);
+	zend_string_release(key);
 }
 
 static xdebug_vector *find_stack_for_fiber(zend_fiber_context *fiber)
 {
 	struct xdebug_fiber_entry *entry = NULL;
-	xdebug_str                *key = create_key_for_fiber(fiber);
+	zend_string               *key = create_key_for_fiber(fiber);
 
-	xdebug_hash_find(XG_BASE(fiber_stacks), key->d, key->l, (void*) &entry);
+	xdebug_hash_find(XG_BASE(fiber_stacks), ZSTR_VAL(key), ZSTR_LEN(key), (void*) &entry);
 
-	xdebug_str_free(key);
+	zend_string_release(key);
 
 	return entry->stack;
 }
@@ -1156,6 +1223,10 @@ static void xdebug_fiber_switch_observer(zend_fiber_context *from, zend_fiber_co
 	xdebug_vector *current_stack;
 
 	if (from->status == ZEND_FIBER_STATUS_DEAD) {
+		if (XG_DBG(context).next_stack == find_stack_for_fiber(from)) {
+			XG_DBG(context).next_stack = NULL;
+		}
+
 		remove_stack_for_fiber(from);
 	}
 	if (to->status == ZEND_FIBER_STATUS_INIT) {
@@ -1169,6 +1240,7 @@ static void xdebug_fiber_switch_observer(zend_fiber_context *from, zend_fiber_co
 		add_fiber_main(to);
 	}
 }
+/***************************************************************************/
 #endif
 
 #ifdef __linux__
@@ -1194,6 +1266,11 @@ int read_systemd_private_tmp_directory(char **private_tmp)
 
 	/* Read contents and split in lines */
 	bytes_read = fread(buffer, 1, sizeof(buffer), mountinfo_fd);
+	if (!bytes_read) {
+		fclose(mountinfo_fd);
+		return retval;
+	}
+
 	lines = xdebug_arg_ctor();
 	xdebug_explode("\n", buffer, lines, -1);
 
@@ -1241,11 +1318,20 @@ void xdebug_base_minit(INIT_FUNC_ARGS)
 	xdebug_old_error_cb = zend_error_cb;
 	xdebug_new_error_cb = xdebug_error_cb;
 
+#if PHP_VERSION_ID >= 80100
+	/* User Code Functions */
+	zend_observer_fcall_register(xdebug_observer_init);
+#endif
+
+	/* Include, Require, Eval */
 	xdebug_old_execute_ex = zend_execute_ex;
 	zend_execute_ex = xdebug_execute_ex;
 
+#if PHP_VERSION_ID < 80200
+	/* Internal Functions, since 8.2 they're also observed */
 	xdebug_old_execute_internal = zend_execute_internal;
 	zend_execute_internal = xdebug_execute_internal;
+#endif
 
 	XG_BASE(error_reporting_override) = 0;
 	XG_BASE(error_reporting_overridden) = 0;
@@ -1258,6 +1344,10 @@ void xdebug_base_minit(INIT_FUNC_ARGS)
 	XG_BASE(private_tmp) = NULL;
 #ifdef __linux__
 	read_systemd_private_tmp_directory(&XG_BASE(private_tmp));
+
+	XG_BASE(control_socket_path) = NULL;
+	XG_BASE(control_socket_fd) = 0;
+	XG_BASE(control_socket_last_trigger) = 0;
 #endif
 }
 
@@ -1308,15 +1398,18 @@ void xdebug_base_rinit()
 	XG_BASE(last_exception_trace) = NULL;
 
 	/* Initialize start time */
-	if (XDEBUG_MODE_IS(XDEBUG_MODE_TRACING) || XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
-		XG_BASE(start_nanotime) = xdebug_get_nanotime();
-	} else {
-		XG_BASE(start_nanotime) = 0;
-	}
+	XG_BASE(start_nanotime) = xdebug_get_nanotime();
 
 	XG_BASE(in_var_serialisation) = 0;
 	zend_ce_closure->serialize = xdebug_closure_serialize_deny_wrapper;
 
+#ifdef __linux__
+	/* Set-up Control Socket */
+	if (XINI_BASE(control_socket_granularity) != XDEBUG_CONTROL_SOCKET_OFF) {
+		xdebug_control_socket_setup();
+	}
+#endif
+
 	/* Signal that we're in a request now */
 	XG_BASE(in_execution) = 1;
 
@@ -1364,6 +1457,11 @@ void xdebug_base_post_deactivate()
 	XG_BASE(filters_code_coverage) = NULL;
 
 	xdebug_base_overloaded_functions_restore();
+
+#ifdef __linux__
+	/* Close Down Control Socket */
+	xdebug_control_socket_teardown();
+#endif
 }
 
 void xdebug_base_rshutdown()
