@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2023 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,8 +12,6 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   +----------------------------------------------------------------------+
  */
 
 #include "php_xdebug.h"
@@ -90,26 +88,26 @@ void xdebug_coverage_function_dtor(void *data)
 static void xdebug_func_format(char *buffer, size_t buffer_size, xdebug_func *func)
 {
 	if (func->type == XFUNC_NORMAL) {
-		int len = strlen(func->function);
+		int len = ZSTR_LEN(func->function);
 
 		if (len + 1 > buffer_size) {
 			goto error;
 		}
-		memcpy(buffer, func->function, len);
+		memcpy(buffer, ZSTR_VAL(func->function), len);
 		buffer[len] = '\0';
 		return;
 	}
 
 	if (func->type == XFUNC_MEMBER) {
-		int func_len = strlen(func->function);
-		int len = ZSTR_LEN(func->class_name) + 2 + func_len;
+		int func_len = ZSTR_LEN(func->function);
+		int len = ZSTR_LEN(func->object_class) + 2 + func_len;
 
 		if (len + 1 > buffer_size) {
 			goto error;
 		}
-		memcpy(buffer, ZSTR_VAL(func->class_name), ZSTR_LEN(func->class_name));
-		memcpy(buffer + ZSTR_LEN(func->class_name), "->", 2);
-		memcpy(buffer + ZSTR_LEN(func->class_name) + 2, func->function, func_len);
+		memcpy(buffer, ZSTR_VAL(func->object_class), ZSTR_LEN(func->object_class));
+		memcpy(buffer + ZSTR_LEN(func->object_class), "->", 2);
+		memcpy(buffer + ZSTR_LEN(func->object_class) + 2, ZSTR_VAL(func->function), func_len);
 		buffer[len] = '\0';
 		return;
 	}
@@ -121,25 +119,31 @@ error:
 
 static void xdebug_build_fname_from_oparray(xdebug_func *tmp, zend_op_array *opa)
 {
-	int closure = 0;
+	int wrapped = 0;
 
 	memset(tmp, 0, sizeof(xdebug_func));
 
 	if (opa->function_name) {
 		if (opa->fn_flags & ZEND_ACC_CLOSURE) {
-			tmp->function = xdebug_wrap_closure_location_around_function_name(opa, STR_NAME_VAL(opa->function_name));
-			closure = 1;
+			tmp->function = xdebug_wrap_closure_location_around_function_name(opa, opa->function_name);
+			wrapped = 1;
+		} else if (
+			opa->fn_flags & ZEND_ACC_TRAIT_CLONE
+			|| (opa->scope && opa->scope->ce_flags & ZEND_ACC_TRAIT)
+		) {
+			tmp->function = xdebug_wrap_location_around_function_name("trait-method", opa, opa->function_name);
+			wrapped = 1;
 		} else {
-			tmp->function = xdstrdup(STR_NAME_VAL(opa->function_name));
+			tmp->function = zend_string_copy(opa->function_name);
 		}
 	} else {
-		tmp->function = xdstrdup("{main}");
+		tmp->function = ZSTR_INIT_LITERAL("{main}", false);
 		tmp->type = XFUNC_MAIN;
 	}
 
-	if (opa->scope && !closure) {
+	if (opa->scope && !wrapped) {
 		tmp->type = XFUNC_MEMBER;
-		tmp->class_name = zend_string_copy(opa->scope->name);
+		tmp->object_class = zend_string_copy(opa->scope->name);
 	} else {
 		tmp->type = XFUNC_NORMAL;
 	}
@@ -154,11 +158,14 @@ static void xdebug_print_opcode_info(zend_execute_data *execute_data, const zend
 
 	xdebug_build_fname_from_oparray(&func_info, op_array);
 	xdebug_func_format(function_name, sizeof(function_name), &func_info);
-	if (func_info.class_name) {
-		zend_string_release(func_info.class_name);
+	if (func_info.object_class) {
+		zend_string_release(func_info.object_class);
+	}
+	if (func_info.scope_class) {
+		zend_string_release(func_info.scope_class);
 	}
 	if (func_info.function) {
-		xdfree(func_info.function);
+		zend_string_release(func_info.function);
 	}
 
 	xdebug_branch_info_mark_reached(op_array->filename, function_name, op_array, opnr);
@@ -253,10 +260,6 @@ static void prefill_from_opcode(zend_string *filename, zend_op opcode, int deadc
 		opcode.opcode != ZEND_EXT_NOP &&
 		opcode.opcode != ZEND_RECV &&
 		opcode.opcode != ZEND_RECV_INIT
-#if PHP_VERSION_ID < 70400
-		&& opcode.opcode != ZEND_VERIFY_ABSTRACT_CLASS
-		&& opcode.opcode != ZEND_ADD_INTERFACE
-#endif
 		&& opcode.opcode != ZEND_OP_DATA
 		&& opcode.opcode != ZEND_TICKS
 		&& opcode.opcode != ZEND_FAST_CALL
@@ -291,11 +294,13 @@ static int xdebug_find_jumps(zend_op_array *opa, unsigned int position, size_t *
 		*jump_count = 2;
 		return 1;
 
+#if PHP_VERSION_ID < 80200
 	} else if (opcode.opcode == ZEND_JMPZNZ) {
 		jumps[0] = XDEBUG_ZNODE_JMP_LINE(opcode.op2, position, base_address);
 		jumps[1] = position + ((int32_t) opcode.extended_value / (int32_t) sizeof(zend_op));
 		*jump_count = 2;
 		return 1;
+#endif
 
 	} else if (opcode.opcode == ZEND_FE_FETCH_R || opcode.opcode == ZEND_FE_FETCH_RW) {
 		jumps[0] = position + 1;
@@ -312,13 +317,8 @@ static int xdebug_find_jumps(zend_op_array *opa, unsigned int position, size_t *
 	} else if (opcode.opcode == ZEND_CATCH) {
 		*jump_count = 2;
 		jumps[0] = position + 1;
-#if PHP_VERSION_ID >= 70300
 		if (!(opcode.extended_value & ZEND_LAST_CATCH)) {
 			jumps[1] = XDEBUG_ZNODE_JMP_LINE(opcode.op2, position, base_address);
-#else
-		if (!opcode.result.num) {
-			jumps[1] = position + (opcode.extended_value / sizeof(zend_op));
-#endif
 			if (jumps[1] == jumps[0]) {
 				jumps[1] = XDEBUG_JMP_NOT_SET;
 				*jump_count = 1;
@@ -348,18 +348,14 @@ static int xdebug_find_jumps(zend_op_array *opa, unsigned int position, size_t *
 		opcode.opcode == ZEND_GENERATOR_RETURN ||
 		opcode.opcode == ZEND_EXIT ||
 		opcode.opcode == ZEND_THROW ||
-#if PHP_VERSION_ID >= 80000
 		opcode.opcode == ZEND_MATCH_ERROR ||
-#endif
 		opcode.opcode == ZEND_RETURN
 	) {
 		jumps[0] = XDEBUG_JMP_EXIT;
 		*jump_count = 1;
 		return 1;
 	} else if (
-# if PHP_VERSION_ID >= 80000
 		opcode.opcode == ZEND_MATCH ||
-# endif
 		opcode.opcode == ZEND_SWITCH_LONG ||
 		opcode.opcode == ZEND_SWITCH_STRING
 	) {
@@ -367,11 +363,7 @@ static int xdebug_find_jumps(zend_op_array *opa, unsigned int position, size_t *
 		HashTable *myht;
 		zval *val;
 
-# if PHP_VERSION_ID >= 70300
 		array_value = RT_CONSTANT(&opa->opcodes[position], opcode.op2);
-# else
-		array_value = RT_CONSTANT_EX(opa->literals, opcode.op2);
-# endif
 		myht = Z_ARRVAL_P(array_value);
 
 		/* All 'case' statements */
@@ -386,15 +378,11 @@ static int xdebug_find_jumps(zend_op_array *opa, unsigned int position, size_t *
 		jumps[*jump_count] = position + (opcode.extended_value / sizeof(zend_op));
 		(*jump_count)++;
 
-# if PHP_VERSION_ID >= 80000
 		if (opcode.opcode != ZEND_MATCH) {
-# endif
 			/* The 'next' opcode */
 			jumps[*jump_count] = position + 1;
 			(*jump_count)++;
-# if PHP_VERSION_ID >= 80000
 		}
-# endif
 
 		return 1;
 	}
@@ -556,17 +544,30 @@ static void prefill_from_oparray(zend_string *filename, zend_op_array *op_array)
 		xdebug_build_fname_from_oparray(&func_info, op_array);
 		xdebug_func_format(function_name, sizeof(function_name), &func_info);
 
-		if (func_info.class_name) {
-			zend_string_release(func_info.class_name);
+		if (func_info.object_class) {
+			zend_string_release(func_info.object_class);
+		}
+		if (func_info.scope_class) {
+			zend_string_release(func_info.scope_class);
 		}
 		if (func_info.function) {
-			xdfree(func_info.function);
+			zend_string_release(func_info.function);
 		}
 
 		xdebug_branch_post_process(op_array, branch_info);
 		xdebug_branch_find_paths(branch_info);
 		xdebug_branch_info_add_branches_and_paths(filename, (char*) function_name, branch_info);
 	}
+
+#if PHP_VERSION_ID >= 80100
+	if (!op_array->num_dynamic_func_defs) {
+		return;
+	}
+
+	for (i = 0; i < op_array->num_dynamic_func_defs; i++) {
+		prefill_from_oparray(filename, op_array->dynamic_func_defs[i]);
+	}
+#endif
 }
 
 static int prefill_from_function_table(zend_op_array *opa)
@@ -631,22 +632,28 @@ static void xdebug_prefill_code_coverage(zend_op_array *op_array)
 void xdebug_code_coverage_start_of_function(zend_op_array *op_array, char *function_name)
 {
 	xdebug_path *path = xdebug_path_new(NULL);
+	int orig_size = XG_COV(branches).size;
 
 	xdebug_prefill_code_coverage(op_array);
-	xdebug_path_info_add_path_for_level(XG_COV(paths_stack), path, XG_BASE(level));
+	xdebug_path_info_add_path_for_level(XG_COV(paths_stack), path, XDEBUG_VECTOR_COUNT(XG_BASE(stack)));
 
-	if (XG_COV(branches).size == 0 || XG_BASE(level) >= XG_COV(branches).size) {
-		XG_COV(branches).size = XG_BASE(level) + 32;
+	if (orig_size == 0 || XDEBUG_VECTOR_COUNT(XG_BASE(stack)) >= orig_size) {
+		size_t i = 0;
+
+		XG_COV(branches).size = XDEBUG_VECTOR_COUNT(XG_BASE(stack)) + 32;
 		XG_COV(branches).last_branch_nr = realloc(XG_COV(branches).last_branch_nr, sizeof(int) * XG_COV(branches.size));
+		for (i = orig_size; i < XG_COV(branches).size; i++) {
+			XG_COV(branches).last_branch_nr[i] = -1;
+		}
 	}
 
-	XG_COV(branches).last_branch_nr[XG_BASE(level)] = -1;
+	XG_COV(branches).last_branch_nr[XDEBUG_VECTOR_COUNT(XG_BASE(stack))] = -1;
 }
 
 void xdebug_code_coverage_end_of_function(zend_op_array *op_array, zend_string *filename, char *function_name)
 {
 	xdebug_str str = XDEBUG_STR_INITIALIZER;
-	xdebug_path *path = xdebug_path_info_get_path_for_level(XG_COV(paths_stack), XG_BASE(level));
+	xdebug_path *path = xdebug_path_info_get_path_for_level(XG_COV(paths_stack), XDEBUG_VECTOR_COUNT(XG_BASE(stack)));
 
 	if (!path || !path->elements) {
 		return;
@@ -686,13 +693,13 @@ PHP_FUNCTION(xdebug_start_code_coverage)
 
 PHP_FUNCTION(xdebug_stop_code_coverage)
 {
-	zend_long cleanup = 1;
+	zend_bool cleanup = 1;
 
 	if (!XDEBUG_MODE_IS(XDEBUG_MODE_COVERAGE)) {
 		RETURN_FALSE;
 	}
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|l", &cleanup) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|b", &cleanup) == FAILURE) {
 		return;
 	}
 	if (!XG_COV(code_coverage_active)) {
@@ -700,10 +707,14 @@ PHP_FUNCTION(xdebug_stop_code_coverage)
 	}
 
 	if (cleanup) {
-		zend_string_release(XG_COV(previous_filename));
+		if (XG_COV(previous_filename)) {
+			zend_string_release(XG_COV(previous_filename));
+		}
 		XG_COV(previous_filename) = NULL;
 		XG_COV(previous_file) = NULL;
-		zend_string_release(XG_COV(previous_mark_filename));
+		if (XG_COV(previous_mark_filename)) {
+			zend_string_release(XG_COV(previous_mark_filename));
+		}
 		XG_COV(previous_mark_filename) = NULL;
 		XG_COV(previous_mark_file) = NULL;
 		xdebug_hash_destroy(XG_COV(code_coverage_info));
@@ -718,15 +729,8 @@ PHP_FUNCTION(xdebug_stop_code_coverage)
 	RETURN_TRUE;
 }
 
-#if PHP_VERSION_ID >= 80000
 static int xdebug_lineno_cmp(Bucket *f, Bucket *s)
 {
-#else
-static int xdebug_lineno_cmp(const void *a, const void *b)
-{
-	Bucket *f = (Bucket *) a;
-	Bucket *s = (Bucket *) b;
-#endif
 	if (f->h < s->h) {
 		return -1;
 	} else if (f->h > s->h) {
@@ -838,6 +842,7 @@ static void add_cc_function(void *ret, xdebug_hash_element *e)
 	xdebug_coverage_function *function = (xdebug_coverage_function*) e->ptr;
 	zval                     *retval = (zval*) ret;
 	zval                     *function_info;
+	zend_string              *trait_scope = NULL;
 
 	XDEBUG_MAKE_STD_ZVAL(function_info);
 	array_init(function_info);
@@ -847,7 +852,14 @@ static void add_cc_function(void *ret, xdebug_hash_element *e)
 		add_paths(function_info, function->branch_info);
 	}
 
-	add_assoc_zval_ex(retval, function->name, HASH_KEY_STRLEN(function->name), function_info);
+	if ((trait_scope = xdebug_get_trait_scope(function->name)) != NULL) {
+		char *with_scope = xdebug_sprintf("%s->%s", ZSTR_VAL(trait_scope), function->name);
+
+		add_assoc_zval_ex(retval, with_scope, strlen(with_scope), function_info);
+	} else {
+		add_assoc_zval_ex(retval, function->name, HASH_KEY_STRLEN(function->name), function_info);
+	}
+
 
 	efree(function_info);
 }
@@ -971,11 +983,14 @@ int xdebug_coverage_execute_ex(function_stack_entry *fse, zend_op_array *op_arra
 		*tmp_function_name = xdstrdup(buffer);
 		xdebug_code_coverage_start_of_function(op_array, *tmp_function_name);
 
-		if (func_info.class_name) {
-			zend_string_release(func_info.class_name);
+		if (func_info.object_class) {
+			zend_string_release(func_info.object_class);
+		}
+		if (func_info.scope_class) {
+			zend_string_release(func_info.scope_class);
 		}
 		if (func_info.function) {
-			xdfree(func_info.function);
+			zend_string_release(func_info.function);
 		}
 		return 1;
 	}
@@ -1028,18 +1043,13 @@ void xdebug_coverage_minit(INIT_FUNC_ARGS)
 	int i;
 
 	/* Get reserved offsets */
-#if PHP_VERSION_ID >= 80000
 	zend_xdebug_cc_run_offset = zend_get_resource_handle(XDEBUG_NAME);
 	zend_xdebug_filter_offset = zend_get_resource_handle(XDEBUG_NAME);
-#else
-	zend_extension dummy_ext;
-	zend_xdebug_cc_run_offset = zend_get_resource_handle(&dummy_ext);
-	zend_xdebug_filter_offset = zend_get_resource_handle(&dummy_ext);
-#endif
 
 	xdebug_register_with_opcode_multi_handler(ZEND_ASSIGN, xdebug_common_override_handler);
 	xdebug_register_with_opcode_multi_handler(ZEND_ASSIGN_DIM, xdebug_common_override_handler);
 	xdebug_register_with_opcode_multi_handler(ZEND_ASSIGN_OBJ, xdebug_common_override_handler);
+	xdebug_register_with_opcode_multi_handler(ZEND_ASSIGN_STATIC_PROP, xdebug_common_override_handler);
 	xdebug_register_with_opcode_multi_handler(ZEND_QM_ASSIGN, xdebug_common_override_handler);
 	xdebug_register_with_opcode_multi_handler(ZEND_INCLUDE_OR_EVAL, xdebug_coverage_include_or_eval_handler);
 
@@ -1099,10 +1109,6 @@ void xdebug_coverage_minit(INIT_FUNC_ARGS)
 	xdebug_set_opcode_handler(ZEND_ISSET_ISEMPTY_PROP_OBJ, xdebug_common_override_handler);
 	xdebug_set_opcode_handler(ZEND_CASE, xdebug_common_override_handler);
 	xdebug_set_opcode_handler(ZEND_DECLARE_LAMBDA_FUNCTION, xdebug_common_override_handler);
-#if PHP_VERSION_ID < 70400
-	xdebug_set_opcode_handler(ZEND_ADD_TRAIT, xdebug_common_override_handler);
-	xdebug_set_opcode_handler(ZEND_BIND_TRAITS, xdebug_common_override_handler);
-#endif
 	xdebug_set_opcode_handler(ZEND_INSTANCEOF, xdebug_common_override_handler);
 	xdebug_set_opcode_handler(ZEND_FAST_RET, xdebug_common_override_handler);
 	xdebug_set_opcode_handler(ZEND_ROPE_ADD, xdebug_common_override_handler);
@@ -1112,10 +1118,8 @@ void xdebug_coverage_minit(INIT_FUNC_ARGS)
 	xdebug_set_opcode_handler(ZEND_GENERATOR_CREATE, xdebug_common_override_handler);
 	xdebug_set_opcode_handler(ZEND_BIND_STATIC, xdebug_common_override_handler);
 	xdebug_set_opcode_handler(ZEND_BIND_LEXICAL, xdebug_common_override_handler);
-#if PHP_VERSION_ID >= 70400
 	xdebug_set_opcode_handler(ZEND_DECLARE_CLASS, xdebug_common_override_handler);
 	xdebug_set_opcode_handler(ZEND_DECLARE_CLASS_DELAYED, xdebug_common_override_handler);
-#endif
 	xdebug_set_opcode_handler(ZEND_SWITCH_STRING, xdebug_switch_handler);
 	xdebug_set_opcode_handler(ZEND_SWITCH_LONG, xdebug_switch_handler);
 
