@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2021 Derick Rethans                               |
+   | Copyright (c) 2002-2023 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -46,10 +46,13 @@
 # define poll WSAPoll
 #endif
 
+#include "Zend/zend_extensions.h"
+
 #include "com.h"
 
 #include "debugger_private.h"
 #include "handler_dbgp.h"
+#include "ip_info.h"
 #include "lib/crc32.h"
 #include "lib/log.h"
 
@@ -134,6 +137,56 @@ void set_keepalive_options(int fd)
 }
 #endif  // !WIN32 && !WINNT
 
+static char* resolve_pseudo_hosts(const char *requested_hostname)
+{
+#if __linux__
+	/* Does it start with 'xdebug://' ? */
+	if (strncmp(requested_hostname, "xdebug://", strlen("xdebug://")) != 0) {
+		return NULL;
+	}
+
+	/* Check for 'gateway' pseudo host */
+	if (strcmp(requested_hostname, "xdebug://gateway") == 0) {
+#if XDEBUG_GATEWAY_SUPPORT
+		char *gateway = xdebug_get_gateway_ip();
+
+		if (!gateway) {
+			xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "GATEWAY", "Could not find network gateway to use for 'gateway' pseudo-host.");
+			return NULL;
+		}
+
+		xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Found 'gateway' pseudo-host, with IP address '%s'.", gateway);
+		return gateway;
+#else
+		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "PSEUDO-GW-NO-SUPPORT", "Pseudo-host: '%s' is not supported on this host.", requested_hostname + strlen("xdebug://"));
+		return NULL;
+# endif
+	}
+
+	/* Check for 'nameserver' pseudo host */
+	if (strcmp(requested_hostname, "xdebug://nameserver") == 0) {
+#if XDEBUG_NAMESERVER_SUPPORT
+		char *gateway = xdebug_get_private_nameserver();
+
+		if (!gateway) {
+			xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "NAMESERVER", "Could not find a private network nameserver for 'nameserver' pseudo-host.");
+			return NULL;
+		}
+
+		xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Found 'nameserver' pseudo-host, with IP address '%s'.", gateway);
+		return gateway;
+# else
+		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "PSEUDO-NS-NO-SUPPORT", "Pseudo-host: '%s' is not supported on this host.", requested_hostname + strlen("xdebug://"));
+		return NULL;
+# endif
+	}
+
+	xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "PSEUDO-UNKNOWN", "Unknown pseudo-host: '%s', only 'gateway' or 'nameserver' are supported.", requested_hostname + strlen("xdebug://"));
+#endif
+
+	return NULL;
+}
+
 static int xdebug_create_socket(const char *hostname, int dport, int timeout)
 {
 	struct addrinfo            hints;
@@ -332,7 +385,7 @@ static int xdebug_create_socket(const char *hostname, int dport, int timeout)
 		break;
 	}
 
-	/* Free the result returned by getaddrinfo */
+	/* Free the result returned by getaddrinfo, as well as the duplicated hostname */
 	freeaddrinfo(remote);
 
 	/* If we got a socket, set the option "No delay" to true (1) */
@@ -367,15 +420,55 @@ void xdebug_close_socket(int socketfd)
 	SCLOSE(socketfd);
 }
 
+static zval *get_client_discovery_address(char **header)
+{
+	xdebug_arg *headers;
+	int         i;
+	zval       *remote_addr = NULL;
+
+	xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Checking for client discovery headers: '%s'.", XINI_DBG(client_discovery_header));
+
+	headers = xdebug_arg_ctor();
+	xdebug_explode(",", XINI_DBG(client_discovery_header), headers, -1);
+
+	for (i = 0; i < headers->c; ++i) {
+		char *header_name = xdebug_trim(headers->args[i]);
+
+		xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Checking header '%s'.", header_name);
+		remote_addr = zend_hash_str_find(Z_ARRVAL(PG(http_globals)[TRACK_VARS_SERVER]), header_name, HASH_KEY_STRLEN(header_name));
+
+		if (remote_addr) {
+			*header = header_name;
+			xdebug_arg_dtor(headers);
+			return remote_addr;
+		}
+
+		xdfree(header_name);
+	}
+
+	return NULL;
+}
+
 /* Starting the debugger */
 static void xdebug_init_normal_debugger(xdebug_str *connection_attempts)
 {
 	zval *remote_addr = NULL;
 	char *cp = NULL;
 	int   cp_found = 0;
-	const char *header = NULL;
+	char *header = NULL;
 
 	if (!XINI_DBG(discover_client_host)) {
+		char *pseudo_hostname = resolve_pseudo_hosts(XINI_DBG(client_host));
+
+		if (pseudo_hostname) {
+			xdebug_str_add_fmt(connection_attempts, "%s:%ld (through xdebug.client_host/xdebug.client_port, from %s)", pseudo_hostname, XINI_DBG(client_port), XINI_DBG(client_host));
+			xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Connecting to resolved address/port: %s:%ld.", pseudo_hostname, (long int) XINI_DBG(client_port));
+
+			XG_DBG(context).socket = xdebug_create_socket(pseudo_hostname, XINI_DBG(client_port), XINI_DBG(connect_timeout_ms));
+			xdfree(pseudo_hostname);
+			return;
+		}
+
 		xdebug_str_add_fmt(connection_attempts, "%s:%ld (through xdebug.client_host/xdebug.client_port)", XINI_DBG(client_host), XINI_DBG(client_port));
 		xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Connecting to configured address/port: %s:%ld.", XINI_DBG(client_host), (long int) XINI_DBG(client_port));
 
@@ -383,25 +476,8 @@ static void xdebug_init_normal_debugger(xdebug_str *connection_attempts)
 		return;
 	}
 
-	xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Checking remote connect back address.");
-	if (XINI_DBG(client_discovery_header) && XINI_DBG(client_discovery_header)[0]) {
-		header = XINI_DBG(client_discovery_header);
-		xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Checking user configured header '%s'.", XINI_DBG(client_discovery_header));
-
-		remote_addr = zend_hash_str_find(Z_ARRVAL(PG(http_globals)[TRACK_VARS_SERVER]), XINI_DBG(client_discovery_header), HASH_KEY_STRLEN(XINI_DBG(client_discovery_header)));
-	}
-	if (!remote_addr) {
-		header = "HTTP_X_FORWARDED_FOR";
-		xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Checking header 'HTTP_X_FORWARDED_FOR'.");
-
-		remote_addr = zend_hash_str_find(Z_ARRVAL(PG(http_globals)[TRACK_VARS_SERVER]), "HTTP_X_FORWARDED_FOR", HASH_KEY_SIZEOF("HTTP_X_FORWARDED_FOR"));
-	}
-	if (!remote_addr) {
-		header = "REMOTE_ADDR";
-		xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Checking header 'REMOTE_ADDR'.");
-
-		remote_addr = zend_hash_str_find(Z_ARRVAL(PG(http_globals)[TRACK_VARS_SERVER]), "REMOTE_ADDR", HASH_KEY_SIZEOF("REMOTE_ADDR"));
-	}
+	/* Discover client host section */
+	remote_addr = get_client_discovery_address(&header);
 
 	if (remote_addr && strstr(Z_STRVAL_P(remote_addr), "://")) {
 		header = NULL;
@@ -412,7 +488,7 @@ static void xdebug_init_normal_debugger(xdebug_str *connection_attempts)
 
 	if (!remote_addr) {
 		xdebug_str_add_fmt(connection_attempts, "%s:%ld (fallback through xdebug.client_host/xdebug.client_port)", XINI_DBG(client_host), XINI_DBG(client_port));
-		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "HDR", "Could not discover client host through HTTP headers, connecting to configured address/port: %s:%ld. :-|", XINI_DBG(client_host), (long int) XINI_DBG(client_port));
+		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "HDR", "Could not discover client host through HTTP headers, connecting to configured address/port: %s:%ld.", XINI_DBG(client_host), (long int) XINI_DBG(client_port));
 
 		XG_DBG(context).socket = xdebug_create_socket(XINI_DBG(client_host), XINI_DBG(client_port), XINI_DBG(connect_timeout_ms));
 		return;
@@ -427,12 +503,13 @@ static void xdebug_init_normal_debugger(xdebug_str *connection_attempts)
 
 	xdebug_str_add_fmt(connection_attempts, "%s:%ld (from %s HTTP header)", Z_STRVAL_P(remote_addr), XINI_DBG(client_port), header);
 	xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Client host discovered through HTTP header, connecting to %s:%ld.", Z_STRVAL_P(remote_addr), (long int) XINI_DBG(client_port));
+	xdfree(header);
 
 	XG_DBG(context).socket = xdebug_create_socket(Z_STRVAL_P(remote_addr), XINI_DBG(client_port), XINI_DBG(connect_timeout_ms));
 
 	if (XG_DBG(context).socket < 0) {
 		xdebug_str_add_fmt(connection_attempts, ", %s:%ld (fallback through xdebug.client_host/xdebug.client_port)", XINI_DBG(client_host), XINI_DBG(client_port));
-		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "CON", "Could not connect to client host discovered through HTTP headers, connecting to configured address/port: %s:%ld. :-|", XINI_DBG(client_host), (long int) XINI_DBG(client_port));
+		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "CON", "Could not connect to client host discovered through HTTP headers, connecting to configured address/port: %s:%ld.", XINI_DBG(client_host), (long int) XINI_DBG(client_port));
 
 		XG_DBG(context).socket = xdebug_create_socket(XINI_DBG(client_host), XINI_DBG(client_port), XINI_DBG(connect_timeout_ms));
 	}
@@ -451,8 +528,8 @@ static void xdebug_init_cloud_debugger(const char *cloud_id)
 
 	host = xdebug_sprintf("%c.cloud.xdebug.com", (crc & 0x0f) + 'a');
 
-	xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Connecting to configured address/port: %s:%ld.", host, 9020L);
-	XG_DBG(context).socket = xdebug_create_socket(host, 9020, XINI_DBG(connect_timeout_ms));
+	xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Connecting to configured address/port: %s:%ld.", host, XDEBUG_CLOUD_PORT);
+	XG_DBG(context).socket = xdebug_create_socket(host, XDEBUG_CLOUD_PORT, XINI_DBG(connect_timeout_ms));
 
 	xdfree(host);
 }
@@ -475,6 +552,30 @@ static int ide_key_is_cloud_id()
 	return 1;
 }
 
+static void warn_if_opcache_is_loaded_after_xdebug()
+{
+	bool xdebug_loaded = false;
+	zend_llist_element *ext_ptr = zend_extensions.head;
+
+	do
+	{
+		zend_extension *zext = (zend_extension *)ext_ptr->data;
+
+		if (strcmp(zext->name, "Xdebug") == 0) {
+			xdebug_loaded = true;
+		}
+
+		if (strcmp(zext->name, "Zend OPcache") == 0) {
+			if (xdebug_loaded) {
+				xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "OPCACHE", "Debugger is not working optimally, as Xdebug is loaded before Zend OPcache");
+			}
+			return;
+		}
+
+		ext_ptr = ext_ptr->next;
+	} while (ext_ptr != NULL);
+}
+
 static void xdebug_init_debugger()
 {
 	xdebug_str *connection_attempts = xdebug_str_new();
@@ -482,6 +583,8 @@ static void xdebug_init_debugger()
 	/* Get handler from mode */
 	XG_DBG(context).handler = &xdebug_handler_dbgp;
 
+	warn_if_opcache_is_loaded_after_xdebug();
+
 	if (strcmp(XINI_DBG(cloud_id), "") != 0) {
 		xdebug_init_cloud_debugger(XINI_DBG(cloud_id));
 		XG_DBG(context).host_type = XDEBUG_CLOUD;
@@ -495,12 +598,12 @@ static void xdebug_init_debugger()
 
 	/* Check whether we're connected, or why not */
 	if (XG_DBG(context).socket >= 0) {
-		xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Connected to debugging client: %s. :-)", connection_attempts->d);
+		xdebug_log(XLOG_CHAN_DEBUG, XLOG_INFO, "Connected to debugging client: %s.", connection_attempts->d);
 		xdebug_mark_debug_connection_pending();
 
 		if (!XG_DBG(context).handler->remote_init(&(XG_DBG(context)), XDEBUG_REQ)) {
 			/* The request could not be started, ignore it then */
-			xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_ERR, "SES-INIT", "The debug session could not be started. Tried: %s. :-(", connection_attempts->d);
+			xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_ERR, "SES-INIT", "The debug session could not be started. Tried: %s.", connection_attempts->d);
 		} else {
 			/* All is well, turn off script time outs */
 			zend_unset_timeout();
@@ -508,11 +611,11 @@ static void xdebug_init_debugger()
 			zend_set_timeout(EG(timeout_seconds), 0);
 		}
 	} else if (XG_DBG(context).socket == -1) {
-		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_ERR, "NOCON", "Could not connect to debugging client. Tried: %s :-(", connection_attempts->d);
+		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_ERR, "NOCON", "Could not connect to debugging client. Tried: %s.", connection_attempts->d);
 	} else if (XG_DBG(context).socket == -2) {
-		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_ERR, "TIMEOUT", "Time-out connecting to debugging client, waited: " ZEND_LONG_FMT " ms. Tried: %s :-(", XINI_DBG(connect_timeout_ms), connection_attempts->d);
+		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_ERR, "TIMEOUT", "Time-out connecting to debugging client, waited: " ZEND_LONG_FMT " ms. Tried: %s.", XINI_DBG(connect_timeout_ms), connection_attempts->d);
 	} else if (XG_DBG(context).socket == -3) {
-		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_ERR, "NOPERM", "No permission connecting to debugging client (%s). This could be SELinux related. :-(", connection_attempts->d);
+		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_ERR, "NOPERM", "No permission connecting to debugging client (%s). This could be SELinux related.", connection_attempts->d);
 	}
 
 	xdebug_str_free(connection_attempts);
@@ -531,12 +634,6 @@ void xdebug_restart_debugger()
 	xdebug_init_debugger();
 }
 
-/* Remote connection activation and house keeping */
-int xdebug_is_debug_connection_active()
-{
-	return XG_DBG(remote_connection_enabled);
-}
-
 void xdebug_mark_debug_connection_active()
 {
 	XG_DBG(remote_connection_enabled) = 1;
@@ -638,6 +735,12 @@ static int xdebug_handle_start_session()
 		}
 	}
 
+	/* Make sure that if we have a trigger value configured, we don't start the session unless it matches */
+	if (activate_session && xdebug_lib_has_shared_secret()) {
+		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_INFO, "TRGSEC-LEGACY", "Not activating through legacy method because xdebug.trigger_value is set");
+		activate_session = 0;
+	}
+
 	return activate_session;
 }
 
