@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2021 Derick Rethans                               |
+   | Copyright (c) 2002-2023 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -14,13 +14,14 @@
    +----------------------------------------------------------------------+
  */
 
+#include <errno.h>
 #include <sys/types.h>
 
 #ifndef PHP_WIN32
 #include <unistd.h>
 #endif
 
-#include "php.h"
+#include "lib/php-header.h"
 #include "SAPI.h"
 
 #include "ext/standard/php_string.h"
@@ -95,12 +96,40 @@ const char *xdebug_dbgp_status_strings[7] =
 const char *xdebug_dbgp_reason_strings[4] =
 	{"ok", "error", "aborted", "exception"};
 
+#define XDEBUG_ERROR_OK                              0
+#define XDEBUG_ERROR_PARSE                           1
+#define XDEBUG_ERROR_DUP_ARG                         2
+#define XDEBUG_ERROR_INVALID_ARGS                    3
+#define XDEBUG_ERROR_UNIMPLEMENTED                   4
+#define XDEBUG_ERROR_COMMAND_UNAVAILABLE             5
+
+#define XDEBUG_ERROR_CANT_OPEN_FILE                100
+#define XDEBUG_ERROR_STREAM_REDIRECT_FAILED        101 /* unused */
+
+#define XDEBUG_ERROR_BREAKPOINT_NOT_SET            200
+#define XDEBUG_ERROR_BREAKPOINT_TYPE_NOT_SUPPORTED 201
+#define XDEBUG_ERROR_BREAKPOINT_INVALID            202
+#define XDEBUG_ERROR_BREAKPOINT_NO_CODE            203
+#define XDEBUG_ERROR_BREAKPOINT_INVALID_STATE      204
+#define XDEBUG_ERROR_NO_SUCH_BREAKPOINT            205
+#define XDEBUG_ERROR_EVALUATING_CODE               206
+#define XDEBUG_ERROR_INVALID_EXPRESSION            207 /* unused */
+
+#define XDEBUG_ERROR_PROPERTY_NON_EXISTENT         300
+#define XDEBUG_ERROR_PROPERTY_NON_EXISTANT         300 /* compatibility typo */
+#define XDEBUG_ERROR_STACK_DEPTH_INVALID           301
+#define XDEBUG_ERROR_CONTEXT_INVALID               302 /* unused */
+
+#define XDEBUG_ERROR_PROFILING_NOT_STARTED         800
+
+#define XDEBUG_ERROR_ENCODING_NOT_SUPPORTED        900
+
 typedef struct {
 	int         code;
 	const char *message;
 } xdebug_error_entry;
 
-xdebug_error_entry xdebug_error_codes[24] = {
+static xdebug_error_entry xdebug_error_codes[24] = {
 	{   0, "no error" },
 	{   1, "parse error in command" },
 	{   2, "duplicate arguments in command" },
@@ -127,6 +156,19 @@ xdebug_error_entry xdebug_error_codes[24] = {
 	{  -1, NULL }
 };
 
+static const char *error_message_from_code(int code)
+{
+	xdebug_error_entry *error_entry = &xdebug_error_codes[0];
+
+	while (error_entry->message) {
+		if (code == error_entry->code) {
+			return error_entry->message;
+		}
+		error_entry++;
+	}
+	return NULL;
+}
+
 #define XDEBUG_STR_SWITCH_DECL       char *__switch_variable
 #define XDEBUG_STR_SWITCH(s)         __switch_variable = (s);
 #define XDEBUG_STR_CASE(s)           if (strcmp(__switch_variable, s) == 0) {
@@ -291,6 +333,7 @@ static xdebug_str *make_message(xdebug_con *context, xdebug_xml_node *message)
 static void send_message_ex(xdebug_con *context, xdebug_xml_node *message, int stage)
 {
 	xdebug_str *tmp;
+	size_t      bytes_written;
 
 	/* Sometimes we end up in 'send_message' although the debugging connection
 	 * is already closed. In that case, we early return. */
@@ -299,15 +342,36 @@ static void send_message_ex(xdebug_con *context, xdebug_xml_node *message, int s
 	}
 
 	tmp = make_message(context, message);
-	if ((size_t) SSENDL(context->socket, tmp->d, tmp->l) != tmp->l) {
-		char *sock_error = php_socket_strerror(php_socket_errno(), NULL, 0);
-		char *utime_str = xdebug_nanotime_to_chars(xdebug_get_nanotime(), 6);
 
-		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "SENDERR", "%s: There was a problem sending %zd bytes on socket %d: %s.", utime_str, tmp->l, context->socket, sock_error);
+	bytes_written = SSENDL(context->socket, tmp->d, tmp->l);
+
+	/* Error */
+	if (bytes_written == -1) {
+		int   current_errno = php_socket_errno();
+		char *sock_error = php_socket_strerror(current_errno, NULL, 0);
+
+		if (current_errno == EPIPE) {
+			xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "REMCLOSE", "The debugging client closed the connection on socket %d: %s (error: %d).", context->socket, sock_error, current_errno);
+			xdebug_abort_debugger();
+		} else {
+			xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "SENDERR", "There was a problem sending %zd bytes on socket %d: %s (error: %d).", tmp->l, context->socket, sock_error, current_errno);
+		}
 
 		efree(sock_error);
-		xdfree(utime_str);
+
+		xdebug_str_free(tmp);
+		return;
 	}
+
+	/* Not enough written, we'll allow that for now */
+	if (bytes_written != tmp->l) {
+		char *sock_error = php_socket_strerror(php_socket_errno(), NULL, 0);
+
+		xdebug_log_ex(XLOG_CHAN_DEBUG, XLOG_WARN, "SENDERR", "There was a problem sending %zd bytes on socket %d: only %zd bytes were written: %s.", tmp->l, context->socket, bytes_written, sock_error);
+
+		efree(sock_error);
+	}
+
 	xdebug_str_free(tmp);
 }
 
@@ -452,27 +516,11 @@ static xdebug_str* return_source(zend_string *filename, int begin, int end)
 	}
 }
 
-
-static int check_evaled_code(zend_string *filename_in, char **filename_out)
-{
-	char *end_marker;
-	xdebug_eval_info *ei;
-
-	end_marker = ZSTR_VAL(filename_in) + ZSTR_LEN(filename_in) - strlen("eval()'d code");
-	if (end_marker >= ZSTR_VAL(filename_in) && strcmp("eval()'d code", end_marker) == 0) {
-		if (xdebug_hash_find(XG_DBG(context).eval_id_lookup, ZSTR_VAL(filename_in), ZSTR_LEN(filename_in), (void *) &ei)) {
-			*filename_out = xdebug_sprintf("dbgp://%lu", ei->id);
-			return 1;
-		}
-	}
-	return 0;
-}
-
 static xdebug_xml_node* return_stackframe(int nr)
 {
 	function_stack_entry *fse, *fse_prev;
 	char                 *tmp_fname;
-	char                 *tmp_filename;
+	zend_string          *tmp_filename;
 	xdebug_xml_node      *tmp;
 
 	fse = xdebug_get_stack_frame(nr);
@@ -484,10 +532,10 @@ static xdebug_xml_node* return_stackframe(int nr)
 	xdebug_xml_add_attribute_ex(tmp, "where", xdstrdup(tmp_fname), 0, 1);
 	xdebug_xml_add_attribute_ex(tmp, "level", xdebug_sprintf("%ld", nr), 0, 1);
 	if (fse_prev) {
-		if (check_evaled_code(fse_prev->filename, &tmp_filename)) {
+		if (xdebug_debugger_check_evaled_code(fse_prev->filename, &tmp_filename)) {
 			xdebug_xml_add_attribute_ex(tmp, "type",     xdstrdup("eval"), 0, 1);
-			xdebug_xml_add_attribute_ex(tmp, "filename", tmp_filename, 0, 0);
-			xdfree(tmp_filename);
+			xdebug_xml_add_attribute_ex(tmp, "filename", ZSTR_VAL(tmp_filename), 0, 0);
+			zend_string_release(tmp_filename);
 		} else {
 			xdebug_xml_add_attribute_ex(tmp, "type",     xdstrdup("file"), 0, 1);
 			xdebug_xml_add_attribute_ex(tmp, "filename", xdebug_path_to_url(fse_prev->filename), 0, 1);
@@ -496,13 +544,13 @@ static xdebug_xml_node* return_stackframe(int nr)
 	} else {
 		zend_string *executed_filename = zend_get_executed_filename_ex();
 		int          executed_lineno   = zend_get_executed_lineno();
-		char        *tmp_filename;
+		zend_string *tmp_filename;
 
-		if (check_evaled_code(executed_filename, &tmp_filename)) {
+		if (xdebug_debugger_check_evaled_code(executed_filename, &tmp_filename)) {
 			xdebug_xml_add_attribute_ex(tmp, "type", xdstrdup("eval"), 0, 1);
-			xdebug_xml_add_attribute_ex(tmp, "filename", tmp_filename, 0, 0);
-			xdfree(tmp_filename);
-		} else {
+			xdebug_xml_add_attribute_ex(tmp, "filename", ZSTR_VAL(tmp_filename), 0, 0);
+			zend_string_release(tmp_filename);
+		} else if (executed_filename) {
 			xdebug_xml_add_attribute_ex(tmp, "type", xdstrdup("file"), 0, 1);
 			xdebug_xml_add_attribute_ex(tmp, "filename", xdebug_path_to_url(executed_filename), 0, 1);
 		}
@@ -600,10 +648,10 @@ static void breakpoint_brk_info_add(xdebug_xml_node *xml, xdebug_brk_info *brk_i
 	if (brk_info->exceptionname) {
 		xdebug_xml_add_attribute_ex(xml, "exception", xdstrdup(brk_info->exceptionname), 0, 1);
 	}
-	if (brk_info->temporary) {
-		xdebug_xml_add_attribute(xml, "state", "temporary");
-	} else if (brk_info->disabled) {
+	if (brk_info->disabled) {
 		xdebug_xml_add_attribute(xml, "state", "disabled");
+	} else if (brk_info->temporary) {
+		xdebug_xml_add_attribute(xml, "state", "temporary");
 	} else {
 		xdebug_xml_add_attribute(xml, "state", "enabled");
 	}
@@ -826,6 +874,25 @@ DBGP_FUNC(breakpoint_update)
 	breakpoint_do_action(DBGP_FUNC_PASS_PARAMETERS, BREAKPOINT_ACTION_UPDATE);
 }
 
+static void breakpoint_exists_helper(void *retval, xdebug_hash_element *he, void *key_to_match)
+{
+	int *line_found = (int *) retval;
+
+	xdebug_brk_admin *admin = (xdebug_brk_admin*) he->ptr;
+
+	if (strcmp(admin->key, (const char*) key_to_match) == 0) {
+		*line_found = 1;
+	}
+}
+
+static int line_breakpoint_exists(xdebug_con *context, const char *key)
+{
+	int line_found = 0;
+
+	xdebug_hash_apply_with_argument(context->breakpoint_list, (void *) &line_found, breakpoint_exists_helper, (void*) key);
+
+	return line_found;
+}
 
 static void breakpoint_list_helper(void *xml, xdebug_hash_element *he, void *dummy)
 {
@@ -845,6 +912,23 @@ DBGP_FUNC(breakpoint_list)
 	xdebug_hash_apply_with_argument(context->breakpoint_list, (void *) *retval, breakpoint_list_helper, NULL);
 }
 
+static void warn_if_breakpoint_file_does_not_exist(xdebug_brk_info *brk_info)
+{
+#ifndef WIN32
+	if (brk_info && brk_info->filename && strstr(ZSTR_VAL(brk_info->filename), "://") == NULL) {
+		struct stat buf;
+
+		if (stat(ZSTR_VAL(brk_info->filename), &buf) != 0) {
+			xdebug_log_ex(
+				XLOG_CHAN_DEBUG, XLOG_WARN,
+				"BRKFILE",
+				"Breakpoint file name does not exist: %s (%s).", ZSTR_VAL(brk_info->filename), strerror(errno)
+			);
+		}
+	}
+#endif
+}
+
 DBGP_FUNC(breakpoint_set)
 {
 	xdebug_brk_info      *brk_info;
@@ -888,6 +972,7 @@ DBGP_FUNC(breakpoint_set)
 	if ((strcmp(CMD_OPTION_CHAR('t'), "line") == 0) || (strcmp(CMD_OPTION_CHAR('t'), "conditional") == 0)) {
 		size_t  new_length = 0;
 		char   *tmp_name;
+		function_stack_entry *fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
 
 		if (!CMD_OPTION_SET('n')) {
 			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_INVALID_ARGS);
@@ -897,8 +982,6 @@ DBGP_FUNC(breakpoint_set)
 
 		/* If no filename is given, we use the current one */
 		if (!CMD_OPTION_SET('f')) {
-			function_stack_entry *fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
-
 			if (!fse) {
 				RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_STACK_DEPTH_INVALID);
 			} else {
@@ -922,12 +1005,22 @@ DBGP_FUNC(breakpoint_set)
 			xdfree(tmp_path);
 		}
 
+		warn_if_breakpoint_file_does_not_exist(brk_info);
+
 		/* Perhaps we have a break condition */
 		if (CMD_OPTION_SET('-')) {
 			brk_info->condition = (char*) xdebug_base64_decode((unsigned char*) CMD_OPTION_CHAR('-'), CMD_OPTION_LEN('-'), &new_length);
 		}
 
 		tmp_name = xdebug_sprintf("%s$%lu", ZSTR_VAL(brk_info->filename), brk_info->original_lineno);
+
+		if (line_breakpoint_exists(context, tmp_name)) {
+			xdfree(tmp_name);
+
+			xdebug_brk_info_dtor(brk_info);
+			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_BREAKPOINT_NOT_SET);
+		}
+
 		if (strcmp(CMD_OPTION_CHAR('t'), "line") == 0) {
 			brk_info->id = breakpoint_admin_add(context, XDEBUG_BREAKPOINT_TYPE_LINE, tmp_name);
 		} else {
@@ -943,6 +1036,16 @@ DBGP_FUNC(breakpoint_set)
 				line_breakpoint_resolve_helper(context, lines_list, brk_info);
 			}
 		}
+
+		if (fse) {
+			function_stack_entry *loop_fse = fse;
+			int i;
+			size_t stack_size = XDEBUG_VECTOR_COUNT(XG_BASE(stack));
+
+			for (i = 0; i < stack_size; i++, loop_fse--) {
+				xdebug_debugger_set_has_line_breakpoints(loop_fse);
+			}
+		}
 	} else
 
 	if ((strcmp(CMD_OPTION_CHAR('t'), "call") == 0) || (strcmp(CMD_OPTION_CHAR('t'), "return") == 0)) {
@@ -964,13 +1067,13 @@ DBGP_FUNC(breakpoint_set)
 			brk_info->classname = xdstrdup(CMD_OPTION_CHAR('a'));
 			tmp_name = xdebug_sprintf(
 				"%c/%s::%s",
-				brk_info->function_break_type == XDEBUG_BREAKPOINT_TYPE_CALL ? 'C' : 'R',
+				(brk_info->function_break_type & XDEBUG_BREAKPOINT_TYPE_CALL) ? 'C' : 'R',
 				CMD_OPTION_CHAR('a'), CMD_OPTION_CHAR('m')
 			);
 		} else {
 			tmp_name = xdebug_sprintf(
 				"%c/%s",
-				brk_info->function_break_type == XDEBUG_BREAKPOINT_TYPE_CALL ? 'C' : 'R',
+				(brk_info->function_break_type & XDEBUG_BREAKPOINT_TYPE_CALL) ? 'C' : 'R',
 				CMD_OPTION_CHAR('m')
 			);
 		}
@@ -984,7 +1087,7 @@ DBGP_FUNC(breakpoint_set)
 			xdfree(tmp_name);
 			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_BREAKPOINT_NOT_SET);
 		} else {
-			if (brk_info->function_break_type == XDEBUG_BREAKPOINT_TYPE_CALL) {
+			if (brk_info->function_break_type & XDEBUG_BREAKPOINT_TYPE_CALL) {
 				brk_info->id = breakpoint_admin_add(context, XDEBUG_BREAKPOINT_TYPE_CALL, tmp_name);
 			} else {
 				brk_info->id = breakpoint_admin_add(context, XDEBUG_BREAKPOINT_TYPE_RETURN, tmp_name);
@@ -1026,6 +1129,7 @@ DBGP_FUNC(eval)
 	size_t           new_length = 0;
 	int              res;
 	xdebug_var_export_options *options;
+	zend_string     *return_message;
 
 	if (!CMD_OPTION_SET('-')) {
 		RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_INVALID_ARGS);
@@ -1042,13 +1146,18 @@ DBGP_FUNC(eval)
 	/* base64 decode eval string */
 	eval_string = (char*) xdebug_base64_decode((unsigned char*) CMD_OPTION_CHAR('-'), CMD_OPTION_LEN('-'), &new_length);
 
-	res = xdebug_do_eval(eval_string, &ret_zval);
+	res = xdebug_do_eval(eval_string, &ret_zval, &return_message);
 
 	xdfree(eval_string);
 
 	/* Handle result */
 	if (!res) {
-		RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_EVALUATING_CODE);
+		if (return_message) {
+			RETURN_RESULT_WITH_MESSAGE(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_EVALUATING_CODE, xdebug_sprintf("%s: %s", error_message_from_code(XDEBUG_ERROR_EVALUATING_CODE), ZSTR_VAL(return_message)));
+			zend_string_release(return_message);
+		} else {
+			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_EVALUATING_CODE);
+		}
 	} else {
 		ret_xml = xdebug_get_zval_value_xml_node(NULL, &ret_zval, options);
 		xdebug_xml_add_child(*retval, ret_xml);
@@ -1146,8 +1255,10 @@ DBGP_FUNC(step_over)
 
 	if ((fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack)))) {
 		XG_DBG(context).next_level = fse->level;
+		XG_DBG(context).next_stack = XG_BASE(stack);
 	} else {
 		XG_DBG(context).next_level = 0;
+		XG_DBG(context).next_stack = NULL;
 	}
 }
 
@@ -1314,6 +1425,11 @@ DBGP_FUNC(feature_get)
 			xdebug_xml_add_attribute(*retval, "supported", "1");
 		XDEBUG_STR_CASE_END
 
+		XDEBUG_STR_CASE("breakpoint_include_return_value")
+			xdebug_xml_add_text(*retval, xdebug_sprintf("%ld", XG_DBG(context).breakpoint_include_return_value));
+			xdebug_xml_add_attribute(*retval, "supported", "1");
+		XDEBUG_STR_CASE_END
+
 		XDEBUG_STR_CASE_DEFAULT
 			xdebug_xml_add_text(*retval, xdstrdup(lookup_cmd(CMD_OPTION_CHAR('n')) ? "1" : "0"));
 			xdebug_xml_add_attribute(*retval, "supported", lookup_cmd(CMD_OPTION_CHAR('n')) ? "1" : "0");
@@ -1385,6 +1501,10 @@ DBGP_FUNC(feature_set)
 			XG_DBG(context).breakpoint_details = strtol(CMD_OPTION_CHAR('v'), NULL, 10);
 		XDEBUG_STR_CASE_END
 
+		XDEBUG_STR_CASE("breakpoint_include_return_value")
+			XG_DBG(context).breakpoint_include_return_value = strtol(CMD_OPTION_CHAR('v'), NULL, 10);
+		XDEBUG_STR_CASE_END
+
 		XDEBUG_STR_CASE_DEFAULT
 			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_INVALID_ARGS);
 		XDEBUG_STR_CASE_DEFAULT_END
@@ -1535,7 +1655,7 @@ DBGP_FUNC(property_get)
 		add_var_retval = add_variable_node(*retval, CMD_OPTION_XDEBUG_STR('n'), 1, 0, 0, options);
 		XG_DBG(context).inhibit_notifications = 0;
 
-		if (add_var_retval) {
+		if (add_var_retval == FAILURE) {
 			options->max_data = old_max_data;
 			RETURN_RESULT(XG_DBG(status), XG_DBG(reason), XDEBUG_ERROR_PROPERTY_NON_EXISTENT);
 		}
@@ -1637,7 +1757,7 @@ DBGP_FUNC(property_set)
 
 	/* Do the eval */
 	eval_string = xdebug_sprintf("%s = %s %s", CMD_OPTION_CHAR('n'), cast_as, new_value);
-	res = xdebug_do_eval(eval_string, &ret_zval);
+	res = xdebug_do_eval(eval_string, &ret_zval, NULL);
 
 	/* restore executor state */
 	if (depth > 0) {
@@ -1796,18 +1916,18 @@ static int attach_context_vars(xdebug_xml_node *node, xdebug_var_export_options
 	/* right now, we only have zero, one, or two with one being globals, which
 	 * is always the head of the stack */
 	if (context_id == 1) {
+		zend_string *key;
+
 		/* add super globals */
 		xdebug_lib_set_active_symbol_table(&EG(symbol_table));
 		xdebug_lib_set_active_data(NULL);
-		add_variable_node(node, XDEBUG_STR_WRAP_CHAR("_COOKIE"),  1, 1, 0, options);
-		add_variable_node(node, XDEBUG_STR_WRAP_CHAR("_ENV"),     1, 1, 0, options);
-		add_variable_node(node, XDEBUG_STR_WRAP_CHAR("_FILES"),   1, 1, 0, options);
-		add_variable_node(node, XDEBUG_STR_WRAP_CHAR("_GET"),     1, 1, 0, options);
-		add_variable_node(node, XDEBUG_STR_WRAP_CHAR("_POST"),    1, 1, 0, options);
-		add_variable_node(node, XDEBUG_STR_WRAP_CHAR("_REQUEST"), 1, 1, 0, options);
-		add_variable_node(node, XDEBUG_STR_WRAP_CHAR("_SERVER"),  1, 1, 0, options);
-		add_variable_node(node, XDEBUG_STR_WRAP_CHAR("_SESSION"), 1, 1, 0, options);
-		add_variable_node(node, XDEBUG_STR_WRAP_CHAR("GLOBALS"),  1, 1, 0, options);
+
+		ZEND_HASH_FOREACH_STR_KEY(&EG(symbol_table), key) {
+			if (!HASH_KEY_IS_NUMERIC(key)) {
+				add_variable_node(node, XDEBUG_STR_WRAP_CHAR(HASH_APPLY_KEY_VAL(key)),  1, 1, 0, options);
+			}
+		} ZEND_HASH_FOREACH_END();
+
 		xdebug_lib_set_active_symbol_table(NULL);
 		return 0;
 	}
@@ -1815,21 +1935,17 @@ static int attach_context_vars(xdebug_xml_node *node, xdebug_var_export_options
 	/* add user defined constants */
 	if (context_id == 2) {
 		zend_constant *val;
+		zend_string   *const_name;
 
-		ZEND_HASH_FOREACH_PTR(EG(zend_constants), val) {
+		ZEND_HASH_FOREACH_STR_KEY_PTR(EG(zend_constants), const_name, val) {
 			xdebug_str *tmp_name;
 
-			if (!val->name) {
-				/* skip special constants */
-				continue;
-			}
-
-			if (XDEBUG_ZEND_CONSTANT_MODULE_NUMBER(val) != PHP_USER_CONSTANT) {
+			if (ZEND_CONSTANT_MODULE_NUMBER(val) != PHP_USER_CONSTANT) {
 				/* we're only interested in user defined constants */
 				continue;
 			}
 
-			tmp_name = xdebug_str_create(val->name->val, val->name->len);
+			tmp_name = xdebug_str_create(ZSTR_VAL(const_name), ZSTR_LEN(const_name));
 			add_constant_node(node, tmp_name, &(val->value), options);
 			xdebug_str_free(tmp_name);
 		} ZEND_HASH_FOREACH_END();
@@ -1837,9 +1953,24 @@ static int attach_context_vars(xdebug_xml_node *node, xdebug_var_export_options
 		return 0;
 	}
 
+	/* Add return value special one if set and depth = 0 */
+	if (XG_DBG(context).breakpoint_include_return_value && XG_DBG(current_return_value) && depth == 0) {
+		xdebug_xml_node *tmp_node;
+		xdebug_str *name = xdebug_str_create_from_const_char("$"XDEBUG_RETURN_VALUE_VAR_NAME);
+
+		tmp_node = xdebug_get_zval_value_xml_node(name, XG_DBG(current_return_value), options);
+		xdebug_xml_expand_attribute_value(tmp_node, "facet", "readonly return_value virtual");
+
+		xdebug_xml_add_child(node, tmp_node);
+		xdebug_str_free(name);
+
+		return 0;
+	}
+
 	/* Here the context_id is 0 */
 	if ((fse = xdebug_get_stack_frame(depth))) {
 		function_stack_entry *old_fse = xdebug_get_stack_frame(depth - 1);
+		bool must_add_this = true;
 
 		if (depth > 0) {
 			xdebug_lib_set_active_data(old_fse->execute_data);
@@ -1849,6 +1980,8 @@ static int attach_context_vars(xdebug_xml_node *node, xdebug_var_export_options
 		xdebug_lib_set_active_symbol_table(fse->symbol_table);
 
 		/* Only show vars when they are scanned */
+		xdebug_lib_register_compiled_variables(fse);
+
 		if (fse->declared_vars) {
 			xdebug_hash *tmp_hash;
 
@@ -1865,24 +1998,26 @@ static int attach_context_vars(xdebug_xml_node *node, xdebug_var_export_options
 			xdebug_hash_apply_with_argument(tmp_hash, (void *) node, func, (void *) options);
 
 			/* Zend engine 2 does not give us $this, eval so we can get it */
-			if (!xdebug_hash_find(tmp_hash, "this", 4, (void *) &var_name)) {
-				add_variable_node(node, XDEBUG_STR_WRAP_CHAR("this"), 1, 1, 0, options);
+			if (xdebug_hash_find(tmp_hash, "this", 4, (void *) &var_name)) {
+				must_add_this = false;
 			}
 
 			xdebug_hash_destroy(tmp_hash);
 		}
 
+		if (must_add_this) {
+			add_variable_node(node, XDEBUG_STR_WRAP_CHAR("this"), 1, 1, 0, options);
+		}
+
 		/* Check for static variables and constants, but only if it's a static
 		 * method call as we attach constants and static properties to "this"
 		 * too normally. */
 		if (fse->function.type == XFUNC_STATIC_MEMBER) {
 			zend_class_entry *ce = zend_fetch_class(fse->function.object_class, ZEND_FETCH_CLASS_DEFAULT);
 
-#if PHP_VERSION_ID >= 70400
 			if (ce->type == ZEND_INTERNAL_CLASS || (ce->ce_flags & ZEND_ACC_IMMUTABLE)) {
 				zend_class_init_statics(ce);
 			}
-#endif
 
 			xdebug_var_xml_attach_static_vars(node, options, ce);
 		}
@@ -2021,159 +2156,6 @@ DBGP_FUNC(xcmd_get_executable_lines)
 /*****************************************************************************
 ** Parsing functions
 */
-
-/* {{{ Constants for state machine */
-#define STATE_NORMAL                   0
-#define STATE_QUOTED                   1
-#define STATE_OPT_FOLLOWS              2
-#define STATE_SEP_FOLLOWS              3
-#define STATE_VALUE_FOLLOWS_FIRST_CHAR 4
-#define STATE_VALUE_FOLLOWS            5
-#define STATE_SKIP_CHAR                6
-#define STATE_ESCAPED_CHAR_FOLLOWS     7
-/* }}} */
-
-static void xdebug_dbgp_arg_dtor(xdebug_dbgp_arg *arg)
-{
-	int i;
-
-	for (i = 0; i < 27; i++) {
-		if (arg->value[i]) {
-			xdebug_str_free(arg->value[i]);
-		}
-	}
-	xdfree(arg);
-}
-
-static int xdebug_dbgp_parse_cmd(char *line, char **cmd, xdebug_dbgp_arg **ret_args)
-{
-	xdebug_dbgp_arg *args = NULL;
-	char *ptr;
-	int   state;
-	int   charescaped = 0;
-	char  opt = ' ', *value_begin = NULL;
-
-	args = xdmalloc(sizeof (xdebug_dbgp_arg));
-	memset(args->value, 0, sizeof(args->value));
-	*cmd = NULL;
-
-	/* Find the end of the command, this is always on the first space */
-	ptr = strchr(line, ' ');
-	if (!ptr) {
-		/* No space found. If the line is not empty, return the line
-		 * and assume it only consists of the command name. If the line
-		 * is 0 chars long, we return a failure. */
-		if (strlen(line)) {
-			*cmd = strdup(line);
-			*ret_args = args;
-			return XDEBUG_ERROR_OK;
-		} else {
-			goto parse_error;
-		}
-	} else {
-		/* A space was found, so we copy everything before it
-		 * into the cmd parameter. */
-		*cmd = xdcalloc(1, ptr - line + 1);
-		memcpy(*cmd, line, ptr - line);
-	}
-	/* Now we loop until we find the end of the string, which is the \0
-	 * character */
-	state = STATE_NORMAL;
-	do {
-		ptr++;
-		switch (state) {
-			case STATE_NORMAL:
-				if (*ptr != '-') {
-					goto parse_error;
-				} else {
-					state = STATE_OPT_FOLLOWS;
-				}
-				break;
-			case STATE_OPT_FOLLOWS:
-				opt = *ptr;
-				state = STATE_SEP_FOLLOWS;
-				break;
-			case STATE_SEP_FOLLOWS:
-				if (*ptr != ' ') {
-					goto parse_error;
-				} else {
-					state = STATE_VALUE_FOLLOWS_FIRST_CHAR;
-					value_begin = ptr + 1;
-				}
-				break;
-			case STATE_VALUE_FOLLOWS_FIRST_CHAR:
-				if (*ptr == '"' && opt != '-') {
-					value_begin = ptr + 1;
-					state = STATE_QUOTED;
-				} else {
-					state = STATE_VALUE_FOLLOWS;
-				}
-				break;
-			case STATE_VALUE_FOLLOWS:
-				if ((*ptr == ' ' && opt != '-') || *ptr == '\0') {
-					int opt_index = opt - 'a';
-
-					if (opt == '-') {
-						opt_index = 26;
-					}
-
-					if (!args->value[opt_index]) {
-						args->value[opt_index] = xdebug_str_create(value_begin, ptr - value_begin);
-						state = STATE_NORMAL;
-					} else {
-						goto duplicate_opts;
-					}
-				}
-				break;
-			case STATE_QUOTED:
-				if (*ptr == '\\') {
-					state = STATE_ESCAPED_CHAR_FOLLOWS;
-				} else
-				if (*ptr == '"') {
-					int opt_index = opt - 'a';
-
-					if (charescaped) {
-						charescaped = 0;
-						break;
-					}
-					if (opt == '-') {
-						opt_index = 26;
-					}
-
-					if (!args->value[opt_index]) {
-						int len = ptr - value_begin;
-
-						args->value[opt_index] = xdebug_str_create(value_begin, len);
-						xdebug_stripcslashes(args->value[opt_index]->d, &len);
-						args->value[opt_index]->l = len;
-
-						state = STATE_SKIP_CHAR;
-					} else {
-						goto duplicate_opts;
-					}
-				}
-				break;
-			case STATE_SKIP_CHAR:
-				state = STATE_NORMAL;
-				break;
-			case STATE_ESCAPED_CHAR_FOLLOWS:
-				state = STATE_QUOTED;
-				break;
-
-		}
-	} while (*ptr);
-	*ret_args = args;
-	return XDEBUG_ERROR_OK;
-
-parse_error:
-	*ret_args = args;
-	return XDEBUG_ERROR_PARSE;
-
-duplicate_opts:
-	*ret_args = args;
-	return XDEBUG_ERROR_DUP_ARG;
-}
-
 static int xdebug_dbgp_parse_option(xdebug_con *context, char* line, int flags, xdebug_xml_node *retval)
 {
 	char *cmd = NULL;
@@ -2183,7 +2165,7 @@ static int xdebug_dbgp_parse_option(xdebug_con *context, char* line, int flags,
 	xdebug_xml_node *error;
 
 	xdebug_log(XLOG_CHAN_DEBUG, XLOG_COM, "<- %s", line);
-	res = xdebug_dbgp_parse_cmd(line, (char**) &cmd, (xdebug_dbgp_arg**) &args);
+	res = xdebug_cmd_parse(line, (char**) &cmd, (xdebug_dbgp_arg**) &args);
 
 	/* Add command name to return packet */
 	if (cmd) {
@@ -2248,7 +2230,7 @@ static int xdebug_dbgp_parse_option(xdebug_con *context, char* line, int flags,
 	}
 
 	xdfree(cmd);
-	xdebug_dbgp_arg_dtor(args);
+	xdebug_cmd_arg_dtor(args);
 	return ret;
 }
 
@@ -2456,6 +2438,7 @@ int xdebug_dbgp_init(xdebug_con *context, int mode)
 	context->inhibit_notifications = 0;
 	context->resolved_breakpoints = 0;
 	context->breakpoint_details = 0;
+	context->breakpoint_include_return_value = 0;
 
 	xdebug_mark_debug_connection_active();
 	xdebug_dbgp_cmdloop(context, XDEBUG_CMDLOOP_BAIL);
@@ -2581,11 +2564,10 @@ int xdebug_dbgp_error(xdebug_con *context, int type, char *exception_type, char
 	return 1;
 }
 
-int xdebug_dbgp_break_on_line(xdebug_con *context, xdebug_brk_info *brk, zend_string *filename, int lineno)
+int xdebug_dbgp_break_on_line(xdebug_con *context, xdebug_brk_info *brk, zend_string *orig_filename, int lineno)
 {
-	char *tmp_file      = ZSTR_VAL(filename);
-	int   tmp_file_len  = ZSTR_LEN(filename);
-	int   free_tmp_file = 0;
+	zend_string *resolved_filename = orig_filename;
+	bool         free_eval_filename = false;
 
 	xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "Checking whether to break on %s:%d.", ZSTR_VAL(brk->filename), brk->resolved_lineno);
 
@@ -2594,52 +2576,43 @@ int xdebug_dbgp_break_on_line(xdebug_con *context, xdebug_brk_info *brk, zend_st
 		return 0;
 	}
 
-	xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "I: Current location: %s:%d.", tmp_file, lineno);
+	xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "I: Current location: %s:%d.", ZSTR_VAL(orig_filename), lineno);
 
-	if (is_dbgp_url(brk->filename) && check_evaled_code(filename, &tmp_file)) {
-		free_tmp_file = 1;
-		tmp_file_len = strlen(tmp_file);
-		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "I: Found eval code for '%s': %s.", ZSTR_VAL(filename), tmp_file);
+	if (is_dbgp_url(brk->filename) && xdebug_debugger_check_evaled_code(orig_filename, &resolved_filename)) {
+		free_eval_filename = true;
+		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "I: Found eval code for '%s': %s.", ZSTR_VAL(orig_filename), ZSTR_VAL(resolved_filename));
 	}
 
-	xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "I: Matching breakpoint '%s:%d' against location '%s:%d'.", ZSTR_VAL(brk->filename), brk->resolved_lineno, tmp_file, lineno);
-
-	if (ZSTR_LEN(brk->filename) != tmp_file_len) {
-		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "R: File name length (%d) doesn't match with breakpoint (%zd).", tmp_file_len, ZSTR_LEN(brk->filename));
-
-		if (free_tmp_file) {
-			xdfree(tmp_file);
-		}
-		return 0;
-	}
+	xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "I: Matching breakpoint '%s:%d' against location '%s:%d'.", ZSTR_VAL(brk->filename), brk->resolved_lineno, ZSTR_VAL(resolved_filename), lineno);
 
 	if (brk->resolved_lineno != lineno) {
 		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "R: Line number (%d) doesn't match with breakpoint (%d).", lineno, brk->resolved_lineno);
 
-		if (free_tmp_file) {
-			xdfree(tmp_file);
+		if (free_eval_filename) {
+			zend_string_release(resolved_filename);
 		}
 		return 0;
 	}
 
-	if (strncasecmp(ZSTR_VAL(brk->filename), tmp_file, ZSTR_LEN(brk->filename)) == 0) {
+	if (zend_string_equals(brk->filename, resolved_filename)) {
 		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "F: File names match (%s).", ZSTR_VAL(brk->filename));
 
-		if (free_tmp_file) {
-			xdfree(tmp_file);
+		if (free_eval_filename) {
+			zend_string_release(resolved_filename);
 		}
 		return 1;
 	}
 
-	xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "R: File names (%s) doesn't match with breakpoint (%s).", tmp_file, ZSTR_VAL(brk->filename));
+	xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "R: File names (%s) doesn't match with breakpoint (%s).", ZSTR_VAL(resolved_filename), ZSTR_VAL(brk->filename));
 
-	if (free_tmp_file) {
-		xdfree(tmp_file);
+	if (free_eval_filename) {
+		zend_string_release(resolved_filename);
 	}
+
 	return 0;
 }
 
-int xdebug_dbgp_breakpoint(xdebug_con *context, xdebug_vector *stack, zend_string *filename, long lineno, int type, char *exception, char *code, const char *message, xdebug_brk_info *brk_info)
+int xdebug_dbgp_breakpoint(xdebug_con *context, xdebug_vector *stack, zend_string *filename, long lineno, int type, char *exception, char *code, const char *message, xdebug_brk_info *brk_info, zval *return_value)
 {
 	xdebug_xml_node *response, *error_container;
 
@@ -2660,11 +2633,11 @@ int xdebug_dbgp_breakpoint(xdebug_con *context, xdebug_vector *stack, zend_strin
 
 	error_container = xdebug_xml_node_init("xdebug:message");
 	if (filename) {
-		char *tmp_filename = NULL;
+		zend_string *tmp_filename = NULL;
 
-		if (check_evaled_code(filename, &tmp_filename)) {
-			xdebug_xml_add_attribute_ex(error_container, "filename", tmp_filename, 0, 0);
-			xdfree(tmp_filename);
+		if (xdebug_debugger_check_evaled_code(filename, &tmp_filename)) {
+			xdebug_xml_add_attribute_ex(error_container, "filename", ZSTR_VAL(tmp_filename), 0, 0);
+			zend_string_release(tmp_filename);
 		} else {
 			xdebug_xml_add_attribute_ex(error_container, "filename", xdebug_path_to_url(filename), 0, 1);
 		}
@@ -2683,6 +2656,18 @@ int xdebug_dbgp_breakpoint(xdebug_con *context, xdebug_vector *stack, zend_strin
 	}
 	xdebug_xml_add_child(response, error_container);
 
+	if (XG_DBG(context).breakpoint_include_return_value && return_value) {
+		xdebug_xml_node *return_value_container, *tmp_node;
+		xdebug_var_export_options *options = (xdebug_var_export_options*) context->options;
+
+		return_value_container = xdebug_xml_node_init("xdebug:return_value");
+		tmp_node = xdebug_get_zval_value_xml_node(NULL, return_value, options);
+
+		xdebug_xml_add_child(return_value_container, tmp_node);
+
+		xdebug_xml_add_child(response, return_value_container);
+	}
+
 	if (XG_DBG(context).breakpoint_details && brk_info) {
 		xdebug_xml_node *breakpoint_node = xdebug_xml_node_init("breakpoint");
 		breakpoint_brk_info_add(breakpoint_node, brk_info);
@@ -2698,8 +2683,18 @@ int xdebug_dbgp_breakpoint(xdebug_con *context, xdebug_vector *stack, zend_strin
 		XG_DBG(lasttransid) = NULL;
 	}
 
+	XG_DBG(current_return_value) = return_value;
+	if (XG_DBG(current_return_value)) {
+		Z_TRY_ADDREF_P(XG_DBG(current_return_value));
+	}
+
 	xdebug_dbgp_cmdloop(context, XDEBUG_CMDLOOP_BAIL);
 
+	if (XG_DBG(current_return_value)) {
+		Z_TRY_DELREF_P(XG_DBG(current_return_value));
+	}
+	XG_DBG(current_return_value) = NULL;
+
 	return xdebug_is_debug_connection_active();
 }
 
@@ -2772,9 +2767,8 @@ static void function_breakpoint_resolve_helper(void *rctxt, xdebug_brk_info *brk
 }
 */
 
-static void line_breakpoint_resolve_helper(xdebug_con *context, xdebug_lines_list *lines_list, xdebug_brk_info *brk_info)
+static xdebug_function_lines_map_item* find_smallest_range(xdebug_lines_list *lines_list, xdebug_brk_info *brk_info)
 {
-//	xdebug_func        func;
 	int                             i;
 	xdebug_function_lines_map_item *found_item = NULL;
 	int                             found_item_span = XDEBUG_RESOLVED_SPAN_MAX;
@@ -2795,6 +2789,16 @@ static void line_breakpoint_resolve_helper(xdebug_con *context, xdebug_lines_lis
 		}
 	}
 
+	return found_item;
+}
+
+static void line_breakpoint_resolve_helper(xdebug_con *context, xdebug_lines_list *lines_list, xdebug_brk_info *brk_info)
+{
+	xdebug_function_lines_map_item *found_item = NULL;
+
+	/* Find smallest function span that fits the line nr in brk_info */
+	found_item = find_smallest_range(lines_list, brk_info);
+
 	if (!found_item) {
 		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "R: Could not find any file/line entry in lines list.");
 		return;
@@ -2932,11 +2936,11 @@ int xdebug_dbgp_notification(xdebug_con *context, zend_string *filename, long li
 
 	error_container = xdebug_xml_node_init("xdebug:message");
 	if (filename) {
-		char *tmp_filename = NULL;
+		zend_string *tmp_filename = NULL;
 
-		if (check_evaled_code(filename, &tmp_filename)) {
-			xdebug_xml_add_attribute_ex(error_container, "filename", tmp_filename, 0, 0);
-			xdfree(tmp_filename);
+		if (xdebug_debugger_check_evaled_code(filename, &tmp_filename)) {
+			xdebug_xml_add_attribute_ex(error_container, "filename", ZSTR_VAL(tmp_filename), 0, 0);
+			zend_string_release(tmp_filename);
 		} else {
 			xdebug_xml_add_attribute_ex(error_container, "filename", xdebug_path_to_url(filename), 0, 1);
 		}
@@ -2983,11 +2987,11 @@ int xdebug_dbgp_user_notify(xdebug_con *context, zend_string *filename, long lin
 
 	location_node = xdebug_xml_node_init("xdebug:location");
 	if (filename) {
-		char *tmp_filename = NULL;
+		zend_string *tmp_filename = NULL;
 
-		if (check_evaled_code(filename, &tmp_filename)) {
-			xdebug_xml_add_attribute_ex(location_node, "filename", tmp_filename, 0, 0);
-			xdfree(tmp_filename);
+		if (xdebug_debugger_check_evaled_code(filename, &tmp_filename)) {
+			xdebug_xml_add_attribute_ex(location_node, "filename", ZSTR_VAL(tmp_filename), 0, 0);
+			zend_string_release(tmp_filename);
 		} else {
 			xdebug_xml_add_attribute_ex(location_node, "filename", xdebug_path_to_url(filename), 0, 1);
 		}
@@ -3026,7 +3030,7 @@ int xdebug_dbgp_register_eval_id(xdebug_con *context, function_stack_entry *fse)
 
 	ei = xdcalloc(sizeof(xdebug_eval_info), 1);
 	ei->id = context->eval_id_sequence;
-	ei->contents = zend_string_copy(fse->include_filename);
+	ei->contents = zend_string_copy(fse->function.include_filename);
 	ei->refcount = 2;
 
 	key = create_eval_key_file(fse->filename, fse->lineno);
