@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2021 Derick Rethans                               |
+   | Copyright (c) 2002-2023 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -37,7 +37,7 @@
 
 ZEND_EXTERN_MODULE_GLOBALS(xdebug)
 
-static const char* text_formats[11] = {
+static const char* text_formats[22] = {
 	"\n",
 	"%s: %s in %s on line %d\n",
 	"\nCall Stack:\n",
@@ -48,10 +48,23 @@ static const char* text_formats[11] = {
 	"\n",
 	"  $%s = %s\n",
 	"  $%s = *uninitialized*\n",
-	"SCREAM:  Error suppression ignored for\n"
+	"SCREAM:  Error suppression ignored for\n",
+	NULL,
+	NULL,
+
+	// 13+ (for xdebug_append_printable_stack_from_zval)
+	"\n%sCall Stack:\n",
+	"",
+	"%sThe stack is empty or not available\n",
+	"%s%10.4F %10ld %3d. %s() %s:%d\n",
+	"\n%s",
+	"\n%sNested Exceptions:\n",
+	"", // nested exceptions footer
+	NULL, // alternative to 16 for html only
+	"\t" // indenter
 };
 
-static const char* ansi_formats[11] = {
+static const char* ansi_formats[22] = {
 	"\n",
 	"[1m[31m%s[0m: %s[22m in [31m%s[0m on line [32m%d[0m[22m\n",
 	"\n[1mCall Stack:[22m\n",
@@ -62,10 +75,23 @@ static const char* ansi_formats[11] = {
 	"\n",
 	"  $%s = %s\n",
 	"  $%s = *uninitialized*\n",
-	"[1m[31mSCREAM[0m:  Error suppression ignored for\n"
+	"[1m[31mSCREAM[0m:  Error suppression ignored for\n",
+	NULL,
+	NULL,
+
+	// 13+ (for xdebug_append_printable_stack_from_zval)
+	"\n%s[1mCall Stack:[22m\n",
+	"",
+	"%sThe stack is empty or not available\n",
+	"%s%10.4F %10ld %3d. %s() %s:%d\n",
+	"\n%s",
+	"\n%s[1mNested Exceptions:[22m\n",
+	"", // nested exceptions footer
+	NULL, // alternative to 16 for html only
+	"\t" // indenter
 };
 
-static const char* html_formats[13] = {
+static const char* html_formats[22] = {
 	"<br />\n<font size='1'><table class='xdebug-error xe-%s%s' dir='ltr' border='1' cellspacing='0' cellpadding='1'>\n",
 	"<tr><th align='left' bgcolor='#f57900' colspan=\"5\"><span style='background-color: #cc0000; color: #fce94f; font-size: x-large;'>( ! )</span> %s: %s in %s on line <i>%d</i></th></tr>\n",
 	"<tr><th align='left' bgcolor='#e9b96e' colspan='5'>Call Stack</th></tr>\n<tr><th align='center' bgcolor='#eeeeec'>#</th><th align='left' bgcolor='#eeeeec'>Time</th><th align='left' bgcolor='#eeeeec'>Memory</th><th align='left' bgcolor='#eeeeec'>Function</th><th align='left' bgcolor='#eeeeec'>Location</th></tr>\n",
@@ -78,7 +104,18 @@ static const char* html_formats[13] = {
 	"<tr><td colspan='2' align='right' bgcolor='#eeeeec' valign='top'><pre>$%s&nbsp;=</pre></td><td colspan='3' bgcolor='#eeeeec' valign='top'><i>Undefined</i></td></tr>\n",
 	" )</td><td title='%s' bgcolor='#eeeeec'><a style='color: black' href='%s'>%s<b>:</b>%d</a></td></tr>\n",
 	"<tr><th align='left' bgcolor='#f57900' colspan=\"5\"><span style='background-color: #cc0000; color: #fce94f; font-size: x-large;'>( ! )</span> %s: %s in <a style='color: black' href='%s'>%s</a> on line <i>%d</i></th></tr>\n",
-	"<tr><th align='left' bgcolor='#f57900' colspan=\"5\"><span style='background-color: #cc0000; color: #fce94f; font-size: x-large;'>( ! )</span> SCREAM: Error suppression ignored for</th></tr>\n"
+	"<tr><th align='left' bgcolor='#f57900' colspan=\"5\"><span style='background-color: #cc0000; color: #fce94f; font-size: x-large;'>( ! )</span> SCREAM: Error suppression ignored for</th></tr>\n",
+
+	// 13+ (for xdebug_append_printable_stack_from_zval)
+	"%s<tr><th align='left' bgcolor='#e9b96e' colspan='5'>Call Stack</th></tr>\n<tr><th align='center' bgcolor='#eeeeec'>#</th><th align='left' bgcolor='#eeeeec'>Time</th><th align='left' bgcolor='#eeeeec'>Memory</th><th align='left' bgcolor='#eeeeec'>Function</th><th align='left' bgcolor='#eeeeec'>Location</th></tr>\n",
+	"",
+	"%s<tr><td colspan='5' bgcolor='#eeeeec'>The stack is empty or not available</td></tr>\n",
+	"%s<tr><td bgcolor='#eeeeec' align='center'>%d</td><td bgcolor='#eeeeec' align='center'>%.4F</td><td bgcolor='#eeeeec' align='right'>%ld</td><td bgcolor='#eeeeec'>%s()</td><td title='%s' bgcolor='#eeeeec'><a style='color: black' href='%s'>%s<b>:</b>%d</a></td></tr>\n",
+	"%s<table class='xdebug-error xe-nested' style='width: 80%; margin: 1em' dir='ltr' border='1' cellspacing='0' cellpadding='1'>\n",
+	"%s<tr><th align='left' bgcolor='#e9b96e' colspan='5'>Nested Exceptions</th></tr><tr><td colspan='5'>\n",
+	"</table></tr>\n", // nested exceptions footer
+	"<tr><td bgcolor='#eeeeec' align='center'>%d</td><td bgcolor='#eeeeec' align='center'>%.4F</td><td bgcolor='#eeeeec' align='right'>%ld</td><td bgcolor='#eeeeec'>%s()</td><td title='%s' bgcolor='#eeeeec'>%s<b>:</b>%d</td></tr>\n",
+	"" // indenter (not used for HTML)
 };
 
 static const char** select_formats(int html)
@@ -243,6 +280,32 @@ void xdebug_append_error_description(xdebug_str *str, int html, const char *erro
 	efree(escaped);
 }
 
+static void xdebug_append_error_description_from_object(xdebug_str *str, int html, zval *exception_obj)
+{
+	zval *message, *file, *line;
+	zval dummy;
+
+	if (Z_TYPE_P(exception_obj) != IS_OBJECT || !instanceof_function(Z_OBJCE_P(exception_obj), zend_ce_throwable)) {
+		return;
+	}
+
+	message = zend_read_property(Z_OBJCE_P(exception_obj), Z_OBJ_P(exception_obj), "message", sizeof("message")-1, 1, &dummy);
+	file = zend_read_property(Z_OBJCE_P(exception_obj), Z_OBJ_P(exception_obj), "file", sizeof("file")-1, 1, &dummy);
+	line = zend_read_property(Z_OBJCE_P(exception_obj), Z_OBJ_P(exception_obj), "line", sizeof("line")-1, 1, &dummy);
+
+	if (!message || !file || !line || Z_TYPE_P(message) != IS_STRING || Z_TYPE_P(file) != IS_STRING || Z_TYPE_P(line) != IS_LONG) {
+		return;
+	}
+
+	xdebug_append_error_description(str, html, STR_NAME_VAL(Z_OBJCE_P(exception_obj)->name), Z_STRVAL_P(message), Z_STRVAL_P(file), Z_LVAL_P(line));
+}
+
+static void xdebug_append_sub_header(xdebug_str *str, int html)
+{
+	const char **formats = select_formats(html);
+	xdebug_str_add_fmt(str, formats[17], formats[21]);
+}
+
 static void add_single_value(xdebug_str *str, zval *zv, int html)
 {
 	xdebug_str *tmp_value = NULL;
@@ -271,6 +334,194 @@ static void add_single_value(xdebug_str *str, zval *zv, int html)
 	}
 }
 
+static void zval_from_stack_add_frame_parameters(zval *frame, function_stack_entry *fse, bool params_as_values)
+{
+	unsigned int  j;
+	zval         *params;
+	int           variadic_opened = 0;
+	int           sent_variables = fse->varc;
+
+	if (sent_variables > 0 && fse->var[sent_variables-1].is_variadic && Z_ISUNDEF(fse->var[sent_variables-1].data)) {
+		sent_variables--;
+	}
+
+	XDEBUG_MAKE_STD_ZVAL(params);
+	array_init(params);
+	add_assoc_zval_ex(frame, "params", HASH_KEY_SIZEOF("params"), params);
+
+	for (j = 0; j < sent_variables; j++) {
+
+		if (fse->var[j].is_variadic) {
+			zval *vparams;
+
+			XDEBUG_MAKE_STD_ZVAL(vparams);
+			array_init(vparams);
+
+			if (fse->var[j].name) {
+				add_assoc_zval_ex(params, ZSTR_VAL(fse->var[j].name), ZSTR_LEN(fse->var[j].name), vparams);
+			} else {
+				add_index_zval(params, j, vparams);
+			}
+			efree(params);
+			params = vparams;
+			variadic_opened = 1;
+			continue;
+		}
+
+		if (params_as_values) {
+			/* Named parameters */
+			if (fse->var[j].name && !variadic_opened) {
+				if (Z_TYPE(fse->var[j].data) == IS_UNDEF) {
+					add_assoc_null_ex(params, ZSTR_VAL(fse->var[j].name), ZSTR_LEN(fse->var[j].name));
+				} else {
+					Z_TRY_ADDREF(fse->var[j].data);
+					add_assoc_zval_ex(params, ZSTR_VAL(fse->var[j].name), ZSTR_LEN(fse->var[j].name), &fse->var[j].data);
+				}
+				continue;
+			}
+
+			/* Unnamed or Variadic parameters */
+			if (Z_TYPE(fse->var[j].data) == IS_UNDEF) {
+				add_index_null(params, j - variadic_opened);
+			} else {
+				Z_TRY_ADDREF(fse->var[j].data);
+				add_index_zval(params, j - variadic_opened, &fse->var[j].data);
+			}
+
+			continue;
+		} else {
+			xdebug_str *argument = NULL;
+
+			if (!Z_ISUNDEF(fse->var[j].data)) {
+				argument = xdebug_get_zval_value_line(&fse->var[j].data, 0, NULL);
+			} else {
+				argument = xdebug_str_create_from_char((char*) "???");
+			}
+			if (fse->var[j].name && !variadic_opened && argument) {
+				add_assoc_stringl_ex(params, ZSTR_VAL(fse->var[j].name), ZSTR_LEN(fse->var[j].name), argument->d, argument->l);
+			} else {
+				add_index_stringl(params, j - variadic_opened, argument->d, argument->l);
+			}
+			if (argument) {
+				xdebug_str_free(argument);
+				argument = NULL;
+			}
+		}
+	}
+
+	efree(params);
+}
+
+static void zval_from_stack_add_frame_variables(zval *frame, zend_execute_data *edata, HashTable *symbols, zend_op_array *opa)
+{
+	unsigned int j;
+	zval         variables;
+
+	array_init(&variables);
+
+	add_assoc_zval_ex(frame, "variables", HASH_KEY_SIZEOF("variables"), &variables);
+
+	xdebug_lib_set_active_data(edata);
+	xdebug_lib_set_active_symbol_table(symbols);
+
+	for (j = 0; j < (unsigned int) opa->last_var; j++) {
+		xdebug_str *symbol_name;
+		zval        symbol;
+
+		symbol_name = xdebug_str_create_from_char(opa->vars[j]->val);
+		xdebug_get_php_symbol(&symbol, symbol_name);
+		xdebug_str_free(symbol_name);
+
+		if (Z_TYPE(symbol) == IS_UNDEF) {
+			add_assoc_null_ex(&variables, opa->vars[j]->val, opa->vars[j]->len);
+		} else {
+			add_assoc_zval_ex(&variables, opa->vars[j]->val, opa->vars[j]->len, &symbol);
+		}
+	}
+}
+
+static void zval_from_stack_add_frame(zval *output, function_stack_entry *fse, zend_execute_data *edata, bool add_local_vars, bool params_as_values)
+{
+	zval                 *frame;
+
+	/* Initialize frame array */
+	XDEBUG_MAKE_STD_ZVAL(frame);
+	array_init(frame);
+
+	/* Add data */
+	add_assoc_double_ex(frame, "time", HASH_KEY_SIZEOF("time"), XDEBUG_SECONDS_SINCE_START(fse->nanotime));
+	add_assoc_long_ex(frame, "memory", HASH_KEY_SIZEOF("memory"), fse->memory);
+	if (fse->function.function) {
+		add_assoc_str_ex(frame, "function", HASH_KEY_SIZEOF("function"), zend_string_copy(fse->function.function));
+	}
+	if (fse->function.object_class) {
+		add_assoc_string_ex(frame, "type",     HASH_KEY_SIZEOF("type"),     (char*) (fse->function.type == XFUNC_STATIC_MEMBER ? "static" : "dynamic"));
+		add_assoc_str_ex(frame,    "class",    HASH_KEY_SIZEOF("class"),    zend_string_copy(fse->function.object_class));
+	}
+	add_assoc_str_ex(frame, "file", HASH_KEY_SIZEOF("file"), zend_string_copy(fse->filename));
+	add_assoc_long_ex(frame, "line", HASH_KEY_SIZEOF("line"), fse->lineno);
+
+	zval_from_stack_add_frame_parameters(frame, fse, params_as_values);
+
+	if (add_local_vars && fse->op_array && fse->op_array->vars) {
+		zval_from_stack_add_frame_variables(frame, edata, fse->symbol_table, fse->op_array);
+	}
+
+	if (fse->function.include_filename) {
+		add_assoc_str_ex(frame, "include_filename", HASH_KEY_SIZEOF("include_filename"), zend_string_copy(fse->function.include_filename));
+	}
+
+	add_next_index_zval(output, frame);
+	efree(frame);
+}
+
+static void zval_from_stack(zval *output, bool add_local_vars, bool params_as_values)
+{
+	function_stack_entry *fse, *next_fse;
+	unsigned int          i;
+
+	array_init(output);
+
+	fse = XDEBUG_VECTOR_HEAD(XG_BASE(stack));
+	next_fse = fse + 1;
+
+	for (i = 0; i < XDEBUG_VECTOR_COUNT(XG_BASE(stack)) - 1; i++, fse++, next_fse++) {
+		zval_from_stack_add_frame(output, fse, next_fse->execute_data, add_local_vars, params_as_values);
+	}
+}
+
+/* Helpers for last_exception_trace slots */
+
+static zval *last_exception_find_trace(zend_object *obj)
+{
+	int i;
+
+	for (i = 0; i < XDEBUG_LAST_EXCEPTION_TRACE_SLOTS; i++) {
+		if (obj == XG_DEV(last_exception_trace).obj_ptr[i]) {
+			return &XG_DEV(last_exception_trace).stack_trace[i];
+		}
+	}
+	return NULL;
+}
+
+static zval *last_exception_get_slot(zend_object *obj)
+{
+	int slot = XG_DEV(last_exception_trace).next_slot;
+
+	if (XG_DEV(last_exception_trace).obj_ptr[slot] != NULL) {
+		zval_ptr_dtor(&XG_DEV(last_exception_trace).stack_trace[slot]);
+		XG_DEV(last_exception_trace).obj_ptr[slot] = NULL;
+	}
+
+	XG_DEV(last_exception_trace).obj_ptr[slot] = obj;
+
+	XG_DEV(last_exception_trace).next_slot = (slot + 1 == XDEBUG_LAST_EXCEPTION_TRACE_SLOTS ? 0 : slot + 1);
+
+	return &XG_DEV(last_exception_trace).stack_trace[slot];
+}
+
+/* Formatting variables */
+
 #define XDEBUG_VAR_FORMAT_INITIALISED   0
 #define XDEBUG_VAR_FORMAT_UNINITIALISED 1
 
@@ -364,6 +615,91 @@ static void xdebug_dump_used_var_with_contents(void *htmlq, xdebug_hash_element*
 	zval_ptr_dtor_nogc(&zvar);
 }
 
+void xdebug_append_printable_stack_from_zval(xdebug_str *str, bool indent, zval *trace, int html)
+{
+	const char **formats = select_formats(html);
+	zval        *frame;
+	int          counter = 0;
+
+	xdebug_str_add_fmt(str, formats[13], indent ? formats[21] : ""); // header
+
+	if (!trace || Z_TYPE_P(trace) != IS_ARRAY) {
+		xdebug_str_add_fmt(str, formats[15], indent ? formats[21] : ""); // message
+		xdebug_str_add_const(str, formats[14]); // footer
+		return;
+	}
+
+	ZEND_HASH_FOREACH_VAL_IND(HASH_OF(trace), frame) {
+		zval *time, *memory, *class, *type, *function, *file, *line;
+		char *combined_function;
+
+		counter++;
+
+		if (Z_TYPE_P(frame) != IS_ARRAY) {
+			continue;
+		}
+
+		time = zend_hash_str_find(HASH_OF(frame), "time", 4);
+		memory = zend_hash_str_find(HASH_OF(frame), "memory", 6);
+		class = zend_hash_str_find(HASH_OF(frame), "class", 5);
+		type = zend_hash_str_find(HASH_OF(frame), "type", 4);
+		function = zend_hash_str_find(HASH_OF(frame), "function", 8);
+		file = zend_hash_str_find(HASH_OF(frame), "file", 4);
+		line = zend_hash_str_find(HASH_OF(frame), "line", 4);
+
+		if (!time || !memory || !function || !file || !line) {
+			continue;
+		}
+
+		if (Z_TYPE_P(time) != IS_DOUBLE || Z_TYPE_P(memory) != IS_LONG || Z_TYPE_P(function) != IS_STRING || Z_TYPE_P(file) != IS_STRING || Z_TYPE_P(line) != IS_LONG) {
+			continue;
+		}
+
+		if (class && type && Z_TYPE_P(class) == IS_STRING && Z_TYPE_P(type) == IS_STRING) {
+			combined_function = xdebug_sprintf("%s%s%s", Z_STRVAL_P(class), strcmp(Z_STRVAL_P(type), "static") == 0 ? "::" : "->", Z_STRVAL_P(function));
+		} else {
+			combined_function = xdstrdup(Z_STRVAL_P(function));
+		}
+
+		if (html) {
+			char *formatted_filename;
+			xdebug_format_filename(&formatted_filename, "...%s%n", Z_STR_P(file));
+
+			if (strlen(XINI_LIB(file_link_format)) > 0 && strcmp(Z_STRVAL_P(file), "Unknown") != 0) {
+				char *file_link;
+
+				xdebug_format_file_link(&file_link, Z_STRVAL_P(file), Z_LVAL_P(line));
+				xdebug_str_add_fmt(str, formats[16], formats[21], counter, Z_DVAL_P(time), Z_LVAL_P(memory), combined_function, Z_STRVAL_P(file), file_link, formatted_filename, Z_LVAL_P(line));
+				xdfree(file_link);
+			} else {
+				xdebug_str_add_fmt(str, formats[20], counter, Z_DVAL_P(time), Z_LVAL_P(memory), combined_function, Z_STRVAL_P(file), formatted_filename, Z_LVAL_P(line));
+			}
+
+			xdfree(formatted_filename);
+		} else {
+			xdebug_str_add_fmt(str, formats[16], indent ? formats[21] : "", Z_DVAL_P(time), Z_LVAL_P(memory), counter, combined_function, Z_STRVAL_P(file), Z_LVAL_P(line));
+		}
+		xdfree(combined_function);
+
+	} ZEND_HASH_FOREACH_END();
+
+	xdebug_str_add_const(str, formats[14]); // footer
+}
+
+static void xdebug_append_nested_section_header(xdebug_str *str, bool indent, int html)
+{
+	const char **formats = select_formats(html);
+
+	xdebug_str_add_fmt(str, formats[18], indent ? formats[21] : "");
+}
+
+static void xdebug_append_nested_section_footer(xdebug_str *str, int html)
+{
+	const char **formats = select_formats(html);
+
+	xdebug_str_add_const(str, formats[19]);
+}
+
 void xdebug_append_printable_stack(xdebug_str *str, int html)
 {
 	int                   printed_frames = 0;
@@ -450,14 +786,14 @@ void xdebug_append_printable_stack(xdebug_str *str, int html)
 			xdebug_str_add_literal(str, ")");
 		}
 
-		if (fse->include_filename) {
+		if (fse->function.include_filename) {
 			if (html) {
 				xdebug_str_add_literal(str, "<font color='#00bb00'>'");
-				xdebug_str_add_zstr(str, fse->include_filename);
+				xdebug_str_add_zstr(str, fse->function.include_filename);
 				xdebug_str_add_literal(str, "</font>");
 			} else {
 				xdebug_str_addc(str, '\'');
-				xdebug_str_add_zstr(str, fse->include_filename);
+				xdebug_str_add_zstr(str, fse->function.include_filename);
 				xdebug_str_addc(str, '\'');
 			}
 		}
@@ -504,6 +840,9 @@ void xdebug_append_printable_stack(xdebug_str *str, int html)
 			fse = fse - 1;
 			scope_nr--;
 		}
+
+		xdebug_lib_register_compiled_variables(fse);
+
 		if (fse->declared_vars && fse->declared_vars->size) {
 			xdebug_hash *tmp_hash;
 
@@ -849,36 +1188,71 @@ void xdebug_develop_error_cb(int orig_type, const char *error_filename, const un
 void xdebug_develop_throw_exception_hook(zend_object *exception, zval *file, zval *line, zval *code, char *code_str, zval *message)
 {
 	zend_class_entry *exception_ce = exception->ce;
-	zval *xdebug_message_trace, *previous_exception;
 	char *exception_trace;
 	xdebug_str tmp_str = XDEBUG_STR_INITIALIZER;
-	zval dummy;
 
-	previous_exception = zend_read_property(exception_ce, exception, "previous", sizeof("previous")-1, 1, &dummy);
-	if (previous_exception && Z_TYPE_P(previous_exception) == IS_OBJECT) {
-		xdebug_message_trace = zend_read_property(exception_ce, Z_OBJ_P(previous_exception), "xdebug_message", sizeof("xdebug_message")-1, 1, &dummy);
-		if (xdebug_message_trace && Z_TYPE_P(xdebug_message_trace) != IS_NULL) {
-			xdebug_str_add(&tmp_str, Z_STRVAL_P(xdebug_message_trace), 0);
-		}
-	}
+	zval *z_previous_exception, *z_last_exception_slot, *z_previous_trace;
+	zend_object *previous_exception_obj = exception;
+	zval dummy;
 
 	if (!PG(html_errors)) {
 		xdebug_str_addc(&tmp_str, '\n');
 	}
 	xdebug_append_error_description(&tmp_str, PG(html_errors), STR_NAME_VAL(exception_ce->name), message ? Z_STRVAL_P(message) : "", Z_STRVAL_P(file), Z_LVAL_P(line));
-	xdebug_append_printable_stack(&tmp_str, PG(html_errors));
+
+	z_previous_trace = last_exception_find_trace(exception);
+	if (z_previous_trace) {
+		xdebug_append_printable_stack_from_zval(&tmp_str, false, z_previous_trace, PG(html_errors));
+	} else {
+		xdebug_append_printable_stack(&tmp_str, PG(html_errors));
+	}
+
+	/* Loop over previous exceptions until there are none left */
+	{
+		bool first = true;
+		bool found = false;
+
+		do {
+			z_previous_exception = zend_read_property(exception_ce, previous_exception_obj, "previous", sizeof("previous")-1, 1, &dummy);
+			if (!z_previous_exception || Z_TYPE_P(z_previous_exception) != IS_OBJECT) {
+				break;
+			}
+
+			if (first) {
+				first = false;
+				found = true;
+				xdebug_append_nested_section_header(&tmp_str, true, PG(html_errors));
+			}
+
+			xdebug_append_sub_header(&tmp_str, PG(html_errors));
+			xdebug_append_error_description_from_object(&tmp_str, PG(html_errors), z_previous_exception);
+
+			z_previous_trace = last_exception_find_trace(Z_OBJ_P(z_previous_exception));
+			xdebug_append_printable_stack_from_zval(&tmp_str, true, z_previous_trace, PG(html_errors));
+
+			previous_exception_obj = Z_OBJ_P(z_previous_exception);
+		} while (true);
+
+		if (found) {
+			xdebug_append_nested_section_footer(&tmp_str, PG(html_errors));
+		}
+	}
+
+	/* Remember last stack trace so it can be retrieved in an exception handler through
+	 * xdebug_get_function_stack(['from_exception' => $e]) */
+	z_last_exception_slot = last_exception_get_slot(exception);
+	zval_from_stack(z_last_exception_slot, true, true);
+	zval_from_stack_add_frame(z_last_exception_slot, XDEBUG_VECTOR_TAIL(XG_BASE(stack)), EG(current_execute_data), true, true);
+
 	exception_trace = tmp_str.d;
 
-#if PHP_VERSION_ID >= 80200
-	exception_ce->ce_flags |= ZEND_ACC_ALLOW_DYNAMIC_PROPERTIES;
-#endif
-	zend_update_property_string(exception_ce, exception, "xdebug_message", sizeof("xdebug_message")-1, exception_trace);
-
+	/* Save */
 	if (XG_BASE(last_exception_trace)) {
 		xdfree(XG_BASE(last_exception_trace));
 	}
 	XG_BASE(last_exception_trace) = exception_trace;
 
+	/* Display if expected */
 	if (XINI_DEV(show_ex_trace) || (instanceof_function(exception_ce, zend_ce_error) && XINI_DEV(show_error_trace))) {
 		if (PG(log_errors)) {
 			xdebug_log_stack(STR_NAME_VAL(exception_ce->name), Z_STRVAL_P(message), Z_STRVAL_P(file), Z_LVAL_P(line));
@@ -913,12 +1287,9 @@ PHP_FUNCTION(xdebug_get_stack_depth)
    Returns an array representing the current stack */
 PHP_FUNCTION(xdebug_get_function_stack)
 {
-	function_stack_entry *fse;
-	unsigned int          i;
-	unsigned int          j;
-	zval                 *frame;
-	zval                 *params;
-	int                   variadic_opened = 0;
+	HashTable            *options = NULL;
+	bool                  add_local_vars = false;
+	bool                  params_as_values = false;
 
 	if (!XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
 		php_error(E_WARNING, "Function must be enabled in php.ini by setting 'xdebug.mode' to 'develop'");
@@ -926,85 +1297,45 @@ PHP_FUNCTION(xdebug_get_function_stack)
 		return;
 	}
 
-	array_init(return_value);
+	ZEND_PARSE_PARAMETERS_START(0, 1)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_ARRAY_HT_OR_NULL(options)
+	ZEND_PARSE_PARAMETERS_END();
 
-	fse = XDEBUG_VECTOR_HEAD(XG_BASE(stack));
+	if (options) {
+		zval *value;
 
-	for (i = 0; i < XDEBUG_VECTOR_COUNT(XG_BASE(stack)) - 1; i++, fse++) {
-		int sent_variables = fse->varc;
-
-		if (fse->function.function) {
-			if (strcmp(fse->function.function, "xdebug_get_function_stack") == 0) {
-				return;
-			}
-		}
-
-		if (sent_variables > 0 && fse->var[sent_variables-1].is_variadic && Z_ISUNDEF(fse->var[sent_variables-1].data)) {
-			sent_variables--;
-		}
-
-		/* Initialize frame array */
-		XDEBUG_MAKE_STD_ZVAL(frame);
-		array_init(frame);
-
-		/* Add data */
-		if (fse->function.function) {
-			add_assoc_string_ex(frame, "function", HASH_KEY_SIZEOF("function"), fse->function.function);
-		}
-		if (fse->function.object_class) {
-			add_assoc_string_ex(frame, "type",     HASH_KEY_SIZEOF("type"),     (char*) (fse->function.type == XFUNC_STATIC_MEMBER ? "static" : "dynamic"));
-			add_assoc_str_ex(frame,    "class",    HASH_KEY_SIZEOF("class"),    zend_string_copy(fse->function.object_class));
-		}
-		add_assoc_str_ex(frame, "file", HASH_KEY_SIZEOF("file"), zend_string_copy(fse->filename));
-		add_assoc_long_ex(frame, "line", HASH_KEY_SIZEOF("line"), fse->lineno);
-
-		/* Add parameters */
-		XDEBUG_MAKE_STD_ZVAL(params);
-		array_init(params);
-		add_assoc_zval_ex(frame, "params", HASH_KEY_SIZEOF("params"), params);
-
-		for (j = 0; j < sent_variables; j++) {
-			xdebug_str *argument = NULL;
-
-			if (fse->var[j].is_variadic) {
-				zval *vparams;
-
-				XDEBUG_MAKE_STD_ZVAL(vparams);
-				array_init(vparams);
-
-				if (fse->var[j].name) {
-					add_assoc_zval_ex(params, ZSTR_VAL(fse->var[j].name), ZSTR_LEN(fse->var[j].name), vparams);
-				} else {
-					add_index_zval(params, j, vparams);
-				}
-				efree(params);
-				params = vparams;
-				variadic_opened = 1;
-				continue;
-			}
-			if (!Z_ISUNDEF(fse->var[j].data)) {
-				argument = xdebug_get_zval_value_line(&fse->var[j].data, 0, NULL);
+		value = zend_hash_str_find(options, "from_exception", sizeof("from_exception") - 1);
+		if (value && Z_TYPE_P(value) == IS_OBJECT && instanceof_function(Z_OBJCE_P(value), zend_ce_throwable)) {
+			if (Z_OBJ_P(value) == XG_DEV(last_exception_trace).obj_ptr[0]) {
+				Z_TRY_ADDREF(XG_DEV(last_exception_trace).stack_trace[0]);
+				ZVAL_COPY_VALUE(return_value, &XG_DEV(last_exception_trace).stack_trace[0]);
 			} else {
-				argument = xdebug_str_create_from_char((char*) "???");
+				array_init(return_value);
 			}
-			if (fse->var[j].name && !variadic_opened && argument) {
-				add_assoc_stringl_ex(params, ZSTR_VAL(fse->var[j].name), ZSTR_LEN(fse->var[j].name), argument->d, argument->l);
-			} else {
-				add_index_stringl(params, j - variadic_opened, argument->d, argument->l);
-			}
-			if (argument) {
-				xdebug_str_free(argument);
-				argument = NULL;
+
+			if (
+				(zend_hash_str_find(options, "local_vars", sizeof("local_vars") - 1)) ||
+				(zend_hash_str_find(options, "params_as_values", sizeof("params_as_values") - 1))
+			) {
+				php_error(E_WARNING, "The 'local_vars' or 'params_as_values' options are ignored when used with the 'from_exception' option");
 			}
+
+			return;
 		}
 
-		if (fse->include_filename) {
-			add_assoc_str_ex(frame, "include_filename", HASH_KEY_SIZEOF("include_filename"), zend_string_copy(fse->include_filename));
+		value = zend_hash_str_find(options, "local_vars", sizeof("local_vars") - 1);
+		if (value) {
+			add_local_vars = (Z_TYPE_P(value) == IS_TRUE);
 		}
 
-		add_next_index_zval(return_value, frame);
-		efree(params);
-		efree(frame);
+		value = zend_hash_str_find(options, "params_as_values", sizeof("params_as_values") - 1);
+		if (value) {
+			params_as_values = (Z_TYPE_P(value) == IS_TRUE);
+		}
 	}
+
+	zval_from_stack(return_value, add_local_vars, params_as_values);
+
 }
 /* }}} */
