@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2022 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,24 +12,25 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   +----------------------------------------------------------------------+
  */
 
+#include <errno.h>
 #include <stdlib.h>
 #include <string.h>
+
 #ifndef WIN32
-#include <unistd.h>
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/resource.h>
-#include <sys/file.h>
+# include <unistd.h>
+# include <sys/types.h>
+# include <sys/socket.h>
+# include <sys/resource.h>
+# include <sys/file.h>
 #else
-#define PATH_MAX MAX_PATH
-#include <winsock2.h>
-#include <io.h>
-#include <process.h>
+# define PATH_MAX MAX_PATH
+# include <winsock2.h>
+# include <io.h>
+# include <process.h>
 #endif
+
 #include "php_xdebug.h"
 
 #include "mm.h"
@@ -38,6 +39,7 @@
 
 #include "lib_private.h"
 #include "usefulstuff.h"
+#include "log.h"
 
 #include "ext/standard/php_lcg.h"
 #include "ext/standard/flock_compat.h"
@@ -164,6 +166,34 @@ char* xdebug_strrstr(const char* haystack, const char* needle)
 	return loc;
 }
 
+char *xdebug_trim(const char *str)
+{
+	char *trimmed = NULL, *begin = (char *) str, *end = NULL;
+
+	/* trim leading space */
+	while (isspace((unsigned char) *begin)) {
+		++begin;
+	}
+
+	/* All spaces */
+	if (*begin == '\0') {
+		return xdstrdup("");
+	}
+
+	/* trim trailing space */
+	end = begin + strlen(begin) - 1;
+	while (end > begin && isspace((unsigned char) *end)) {
+		--end;
+	}
+	end++;
+
+	trimmed = xdmalloc(end - begin + 1);
+	memcpy(trimmed, begin, (end - begin));
+	trimmed[(end - begin)] = '\0';
+
+	return trimmed;
+}
+
 /* not all versions of php export this */
 static int xdebug_htoi(char *s)
 {
@@ -283,8 +313,8 @@ char *xdebug_path_to_url(zend_string *fileurl)
 	/* encode the url */
 	encoded_fileurl = xdebug_raw_url_encode(ZSTR_VAL(fileurl), ZSTR_LEN(fileurl), &new_len, 1);
 
-	if (strncmp(ZSTR_VAL(fileurl), "phar://", 7) == 0) {
-		/* ignore, phar is cool */
+	if (strstr(ZSTR_VAL(fileurl), "://") != NULL && strstr(ZSTR_VAL(fileurl), "://") < strstr(ZSTR_VAL(fileurl), "/")) {
+		/* ignore, some form of stream wrapper scheme */
 		tmp = xdstrdup(ZSTR_VAL(fileurl));
 	} else if (ZSTR_VAL(fileurl)[0] != '/' && ZSTR_VAL(fileurl)[0] != '\\' && ZSTR_VAL(fileurl)[1] != ':') {
 		/* convert relative paths */
@@ -309,7 +339,7 @@ char *xdebug_path_to_url(zend_string *fileurl)
 
 	} else if (ZSTR_VAL(fileurl)[1] == '/' || ZSTR_VAL(fileurl)[1] == '\\') {
 		/* convert UNC paths (eg. \\server\sharepath) */
-		/* See http://blogs.msdn.com/ie/archive/2006/12/06/file-uris-in-windows.aspx */
+		/* See https://docs.microsoft.com/en-us/archive/blogs/ie/file-uris-in-windows */
 		tmp = xdebug_sprintf("file:%s", encoded_fileurl);
 	} else if (ZSTR_VAL(fileurl)[0] == '/' || ZSTR_VAL(fileurl)[0] == '\\') {
 		/* convert *nix paths (eg. /path) */
@@ -462,10 +492,11 @@ FILE *xdebug_fopen(char *fname, const char *mode, const char *extension, char **
 }
 #endif
 
+
 int xdebug_format_output_filename(char **filename, char *format, char *script_name)
 {
 	xdebug_str fname = XDEBUG_STR_INITIALIZER;
-	char       cwd[128];
+	char       cwd[MAXPATHLEN];
 
 	while (*format)
 	{
@@ -476,7 +507,7 @@ int xdebug_format_output_filename(char **filename, char *format, char *script_na
 			switch (*format)
 			{
 				case 'c': /* crc32 of the current working directory */
-					if (VCWD_GETCWD(cwd, 127)) {
+					if (VCWD_GETCWD(cwd, MAXPATHLEN - 1)) {
 						xdebug_str_add_fmt(&fname, "%lu", xdebug_crc32(cwd, strlen(cwd)));
 					}
 					break;
@@ -516,18 +547,16 @@ int xdebug_format_output_filename(char **filename, char *format, char *script_na
 				}	break;
 
 				case 't': { /* timestamp (in seconds) */
-					xdebug_str_add(&fname, xdebug_nanotime_to_chars(xdebug_get_nanotime(), 0), 1);
+					xdebug_str_add_fmt(&fname, "%lu", xdebug_get_nanotime() / NANOS_IN_SEC);
 				}	break;
 
 				case 'u': { /* timestamp (in microseconds) */
-					char *char_ptr, *utime_str = xdebug_nanotime_to_chars(xdebug_get_nanotime(), 6);
+					uint64_t nanotime = xdebug_get_nanotime();
 
-					/* Replace . with _ (or should it be nuked?) */
-					char_ptr = strrchr(utime_str, '.');
-					if (char_ptr) {
-						char_ptr[0] = '_';
-					}
-					xdebug_str_add(&fname, utime_str, 1);
+					xdebug_str_add_fmt(&fname, "%lu.%06d",
+						nanotime / NANOS_IN_SEC,
+						(nanotime % NANOS_IN_SEC) / NANOS_IN_MICROSEC
+					);
 				}	break;
 
 				case 'H':   /* $_SERVER['HTTP_HOST'] */
