@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2023 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,15 +12,14 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   +----------------------------------------------------------------------+
  */
-#include "php.h"
+#include "lib/php-header.h"
 #include "ext/standard/php_string.h"
 
 #include "php_xdebug.h"
 #include "tracing_private.h"
 #include "trace_textual.h"
+#include "trace_flamegraph.h"
 #include "trace_computerized.h"
 #include "trace_html.h"
 
@@ -39,11 +38,20 @@ static xdebug_trace_handler_t *xdebug_select_trace_handler(int options)
 		case 0: tmp = &xdebug_trace_handler_textual; break;
 		case 1: tmp = &xdebug_trace_handler_computerized; break;
 		case 2: tmp = &xdebug_trace_handler_html; break;
+		case 3: tmp = &xdebug_trace_handler_flamegraph_cost; break;
+		case 4: tmp = &xdebug_trace_handler_flamegraph_mem; break;
 		default:
 			php_error(E_NOTICE, "A wrong value for xdebug.trace_format was selected (%d), defaulting to the textual format", (int) XINI_TRACE(trace_format));
 			tmp = &xdebug_trace_handler_textual; break;
 	}
 
+	/* Override handler based on options */
+	if (options & XDEBUG_TRACE_OPTION_FLAMEGRAPH_COST) {
+		tmp = &xdebug_trace_handler_flamegraph_cost;
+	}
+	if (options & XDEBUG_TRACE_OPTION_FLAMEGRAPH_MEM) {
+		tmp = &xdebug_trace_handler_flamegraph_mem;
+	}
 	if (options & XDEBUG_TRACE_OPTION_COMPUTERIZED) {
 		tmp = &xdebug_trace_handler_computerized;
 	}
@@ -51,12 +59,16 @@ static xdebug_trace_handler_t *xdebug_select_trace_handler(int options)
 		tmp = &xdebug_trace_handler_html;
 	}
 
+	if (!tmp->init || !tmp->deinit || !tmp->get_filename) {
+		xdebug_log_ex(XLOG_CHAN_TRACE, XLOG_CRIT, "HNDLR", "Broken trace handler for format '%d', missing 'init', 'deinit', or 'get_filename'  handler", options);
+	}
+
 	return tmp;
 }
 
-FILE *xdebug_trace_open_file(char *requested_filename, zend_string *script_filename, long options, char **used_fname)
+xdebug_file *xdebug_trace_open_file(char *requested_filename, zend_string *script_filename, long options)
 {
-	FILE *file;
+	xdebug_file *file = xdebug_file_ctor();
 	char *filename_to_use;
 	char *generated_filename = NULL;
 	char *output_dir = xdebug_lib_get_output_dir(); /* not duplicated */
@@ -69,6 +81,7 @@ FILE *xdebug_trace_open_file(char *requested_filename, zend_string *script_filen
 			xdebug_format_output_filename(&generated_filename, XINI_TRACE(trace_output_name), ZSTR_VAL(script_filename)) <= 0
 		) {
 			/* Invalid or empty xdebug.trace_output_name */
+			xdebug_file_dtor(file);
 			return NULL;
 		}
 
@@ -81,13 +94,13 @@ FILE *xdebug_trace_open_file(char *requested_filename, zend_string *script_filen
 			filename_to_use = xdebug_sprintf("%s%c%s", output_dir, DEFAULT_SLASH, generated_filename);
 		}
 	}
-	if (options & XDEBUG_TRACE_OPTION_APPEND) {
-		file = xdebug_fopen(filename_to_use, "a", (options & XDEBUG_TRACE_OPTION_NAKED_FILENAME) ? NULL : "xt", used_fname);
-	} else {
-		file = xdebug_fopen(filename_to_use, "w", (options & XDEBUG_TRACE_OPTION_NAKED_FILENAME) ? NULL : "xt", used_fname);
-	}
 
-	if (!file) {
+	if (!xdebug_file_open(
+		file,
+		filename_to_use,
+		(options & XDEBUG_TRACE_OPTION_NAKED_FILENAME) ? NULL : "xt",
+		(options & XDEBUG_TRACE_OPTION_APPEND) ? "ab" : "wb"
+	)) {
 		xdebug_log_diagnose_permissions(XLOG_CHAN_TRACE, output_dir, generated_filename);
 	}
 
@@ -106,13 +119,19 @@ static char* xdebug_start_trace(char* fname, zend_string *script_filename, long
 	}
 
 	XG_TRACE(trace_handler) = xdebug_select_trace_handler(options);
-	XG_TRACE(trace_context) = (void*) XG_TRACE(trace_handler)->init(fname, script_filename, options);
+	if (!XG_TRACE(trace_handler)) {
+		return NULL;
+	}
 
+	XG_TRACE(trace_context) = (void*) XG_TRACE(trace_handler)->init(fname, script_filename, options);
 	if (!XG_TRACE(trace_context)) {
 		return NULL;
 	}
 
-	XG_TRACE(trace_handler)->write_header(XG_TRACE(trace_context));
+	if (XG_TRACE(trace_handler)->write_header) {
+		XG_TRACE(trace_handler)->write_header(XG_TRACE(trace_context));
+	}
+
 	return xdstrdup(XG_TRACE(trace_handler)->get_filename(XG_TRACE(trace_context)));
 }
 
@@ -122,7 +141,9 @@ static void xdebug_stop_trace(void)
 		return;
 	}
 
-	XG_TRACE(trace_handler)->write_footer(XG_TRACE(trace_context));
+	if (XG_TRACE(trace_handler)->write_footer) {
+		XG_TRACE(trace_handler)->write_footer(XG_TRACE(trace_context));
+	}
 	XG_TRACE(trace_handler)->deinit(XG_TRACE(trace_context));
 	XG_TRACE(trace_context) = NULL;
 }
@@ -173,8 +194,8 @@ PHP_FUNCTION(xdebug_stop_trace)
 	WARN_AND_RETURN_IF_MODE_IS_NOT(XDEBUG_MODE_TRACING);
 
 	if (!XG_TRACE(trace_context)) {
-		RETVAL_FALSE;
 		php_error(E_NOTICE, "Function trace was not started");
+		RETURN_FALSE;
 	}
 
 	RETVAL_STRING(XG_TRACE(trace_handler)->get_filename(XG_TRACE(trace_context)));
@@ -195,7 +216,6 @@ PHP_FUNCTION(xdebug_get_tracefile_name)
 	RETVAL_STRING(filename);
 }
 
-#if PHP_VERSION_ID >= 70400
 static const char *get_assign_operation(uint32_t extended_value)
 {
 	switch (extended_value) {
@@ -215,15 +235,13 @@ static const char *get_assign_operation(uint32_t extended_value)
 			return "";
 	}
 }
-#endif
 
-static int xdebug_is_static_call(const zend_op *cur_opcode, const zend_op *prev_opcode, const zend_op **found_opcode)
+static int xdebug_is_static_call(const zend_op *first_opcode, const zend_op *cur_opcode, const zend_op *prev_opcode, const zend_op **found_opcode)
 {
 	const zend_op *opcode_ptr;
 
 	opcode_ptr = cur_opcode;
 
-# if PHP_VERSION_ID >= 70400
 	if (
 		(opcode_ptr->opcode == ZEND_ASSIGN_STATIC_PROP) || (opcode_ptr->opcode == ZEND_ASSIGN_STATIC_PROP_REF) ||
 		(opcode_ptr->opcode == ZEND_PRE_INC_STATIC_PROP) || (opcode_ptr->opcode == ZEND_PRE_DEC_STATIC_PROP) ||
@@ -232,9 +250,12 @@ static int xdebug_is_static_call(const zend_op *cur_opcode, const zend_op *prev_
 		*found_opcode = opcode_ptr;
 		return 1;
 	}
-# endif
+
 	while (!(opcode_ptr->opcode == ZEND_EXT_STMT) && !((opcode_ptr->opcode == ZEND_FETCH_STATIC_PROP_W) || (opcode_ptr->opcode == ZEND_FETCH_STATIC_PROP_RW))) {
 		opcode_ptr = opcode_ptr - 1;
+		if (opcode_ptr < first_opcode) {
+			return 0;
+		}
 	}
 	if ((opcode_ptr->opcode == ZEND_FETCH_STATIC_PROP_W) || (opcode_ptr->opcode == ZEND_FETCH_STATIC_PROP_RW)) {
 		*found_opcode = opcode_ptr;
@@ -296,9 +317,11 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 	if (cur_opcode->opcode == ZEND_QM_ASSIGN) {
 		xdebug_str_addc(&name, '$');
 		xdebug_str_add(&name, zend_get_compiled_variable_name(op_array, cur_opcode->result.var)->val, 0);
+
+		return name.d;
 	}
 
-	is_static = xdebug_is_static_call(cur_opcode, prev_opcode, &static_opcode_ptr);
+	is_static = xdebug_is_static_call(op_array->opcodes, cur_opcode, prev_opcode, &static_opcode_ptr);
 	options = xdebug_var_export_options_from_ini();
 	options->no_decoration = 1;
 
@@ -325,13 +348,12 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 		xdebug_str_add_str(&name, zval_value);
 		xdebug_str_free(zval_value);
 	}
-#if PHP_VERSION_ID >= 70400
+
 	if (cur_opcode->opcode >= ZEND_PRE_INC_STATIC_PROP && cur_opcode->opcode <= ZEND_POST_DEC_STATIC_PROP) {
 		zval_value = xdebug_get_zval_value_line(xdebug_get_zval(execute_data, cur_opcode->op1_type, &cur_opcode->op1, &is_var), 0, options);
 		xdebug_str_add_str(&name, zval_value);
 		xdebug_str_free(zval_value);
 	}
-#endif
 
 	/* Scroll back to start of FETCHES */
 	/* FIXME: See whether we can do this unroll looping only once - in is_static() */
@@ -403,10 +425,8 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 	}
 
 	if (
-		(cur_opcode->opcode == ZEND_ASSIGN_OBJ)
-#if PHP_VERSION_ID >= 70400
-		|| (cur_opcode->opcode == ZEND_ASSIGN_OBJ_REF)
-#endif
+		(cur_opcode->opcode == ZEND_ASSIGN_OBJ) ||
+		(cur_opcode->opcode == ZEND_ASSIGN_OBJ_REF)
 	) {
 		if (cur_opcode->op1_type == IS_UNUSED) {
 			xdebug_str_add_literal(&name, "$this");
@@ -415,7 +435,7 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 		xdebug_str_add_literal(&name, "->");
 		xdebug_str_add(&name, Z_STRVAL_P(dimval), 0);
 	}
-#if PHP_VERSION_ID >= 70400
+
 	if (cur_opcode->opcode == ZEND_ASSIGN_STATIC_PROP_REF) {
 		dimval = xdebug_get_zval(execute_data, cur_opcode->op1_type, &cur_opcode->op1, &is_var);
 		xdebug_str_add(&name, Z_STRVAL_P(dimval), 0);
@@ -443,29 +463,6 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 		xdebug_str_add_str(&name, zval_value);
 		xdebug_str_free(zval_value);
 	}
-#else
-	if ((cur_opcode->opcode >= ZEND_ASSIGN_ADD && cur_opcode->opcode <= ZEND_ASSIGN_BW_XOR)
-		|| cur_opcode->opcode == ZEND_ASSIGN_POW
-	) {
-		if (cur_opcode->extended_value == ZEND_ASSIGN_DIM) {
-			zval_value = xdebug_get_zval_value_line(xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var), 0, NULL);
-			xdebug_str_addc(&name, '[');
-			xdebug_str_add_str(&name, zval_value);
-			xdebug_str_addc(&name, ']');
-			xdebug_str_free(zval_value);
-		}
-		if (cur_opcode->extended_value == ZEND_ASSIGN_OBJ) {
-			zval_value = xdebug_get_zval_value_line(xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var), 0, options);
-			if (cur_opcode->op1_type == IS_UNUSED) {
-				xdebug_str_add_literal(&name, "$this->");
-			} else {
-				xdebug_str_add_literal(&name, "->");
-			}
-			xdebug_str_add_str(&name, zval_value);
-			xdebug_str_free(zval_value);
-		}
-	}
-#endif
 
 	if (cur_opcode->opcode == ZEND_ASSIGN_DIM) {
 		if (next_opcode->opcode == ZEND_OP_DATA && cur_opcode->op2_type == IS_UNUSED) {
@@ -479,12 +476,10 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 		}
 	}
 
-#if PHP_VERSION_ID >= 70400
 	if (cur_opcode->opcode == ZEND_ASSIGN_STATIC_PROP) {
 		dimval = xdebug_get_zval(execute_data, cur_opcode->op1_type, &cur_opcode->op1, &is_var);
 		xdebug_str_add(&name, Z_STRVAL_P(dimval), 0);
 	}
-#endif
 
 	xdfree(options->runtime);
 	xdfree(options);
@@ -547,7 +542,6 @@ static int xdebug_common_assign_dim_handler(const char *op, XDEBUG_OPCODE_HANDLE
 			full_varname = tmp_varname;
 
 			val = xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var);
-#if PHP_VERSION_ID >= 70400
 		} else if (cur_opcode->opcode >= ZEND_PRE_INC_STATIC_PROP && cur_opcode->opcode <= ZEND_POST_DEC_STATIC_PROP) {
 			char *tmp_varname;
 
@@ -561,13 +555,10 @@ static int xdebug_common_assign_dim_handler(const char *op, XDEBUG_OPCODE_HANDLE
 			full_varname = tmp_varname;
 
 			val = xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var);
-#endif
 		} else if (
-			(next_opcode->opcode == ZEND_OP_DATA)
-#if PHP_VERSION_ID >= 70400
-			&& (cur_opcode->opcode != ZEND_ASSIGN_OBJ_REF)
-			&& (cur_opcode->opcode != ZEND_ASSIGN_STATIC_PROP_REF)
-#endif
+			(next_opcode->opcode == ZEND_OP_DATA) &&
+			(cur_opcode->opcode != ZEND_ASSIGN_OBJ_REF) &&
+			(cur_opcode->opcode != ZEND_ASSIGN_STATIC_PROP_REF)
 		) {
 			val = xdebug_get_zval_with_opline(execute_data, next_opcode, next_opcode->op1_type, &next_opcode->op1, &is_var);
 		} else if (cur_opcode->opcode == ZEND_QM_ASSIGN) {
@@ -579,22 +570,20 @@ static int xdebug_common_assign_dim_handler(const char *op, XDEBUG_OPCODE_HANDLE
 				const zend_op *referenced_opline = xdebug_find_referenced_opline(execute_data, cur_opcode, 2);
 				right_full_varname = xdebug_find_var_name(execute_data, referenced_opline, NULL);
 			}
-#if PHP_VERSION_ID >= 70400
 		} else if (cur_opcode->opcode == ZEND_ASSIGN_OBJ_REF) {
-			if (cur_opcode->op2_type == IS_CV) {
-				right_full_varname = xdebug_sprintf("$%s", zend_get_compiled_variable_name(op_array, cur_opcode->op2.var)->val);
+			if (next_opcode->op1_type == IS_CV) {
+				right_full_varname = xdebug_sprintf("$%s", zend_get_compiled_variable_name(op_array, next_opcode->op1.var)->val);
 			} else {
 				const zend_op *referenced_opline = xdebug_find_referenced_opline(execute_data, next_opcode, 1);
 				right_full_varname = xdebug_find_var_name(execute_data, referenced_opline, NULL);
 			}
 		} else if (cur_opcode->opcode == ZEND_ASSIGN_STATIC_PROP_REF) {
-			if (cur_opcode->op2_type == IS_CV) {
-				right_full_varname = xdebug_sprintf("$%s", zend_get_compiled_variable_name(op_array, cur_opcode->op2.var)->val);
+			if (next_opcode->op1_type == IS_CV) {
+				right_full_varname = xdebug_sprintf("$%s", zend_get_compiled_variable_name(op_array, next_opcode->op1.var)->val);
 			} else {
 				const zend_op *referenced_opline = xdebug_find_referenced_opline(execute_data, next_opcode, 1);
 				right_full_varname = xdebug_find_var_name(execute_data, referenced_opline, NULL);
 			}
-#endif
 		} else {
 			val = xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var);
 		}
@@ -612,24 +601,10 @@ static int xdebug_common_assign_dim_handler(const char *op, XDEBUG_OPCODE_HANDLE
 
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign, "=");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(qm_assign, "=");
-#if PHP_VERSION_ID >= 70400
 XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_op);
 XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_dim_op);
 XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_obj_op);
 XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_static_prop_op);
-#else
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_add, "+=");
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_sub, "-=");
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_mul, "*=");
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_div, "/=");
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_mod, "%=");
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_sl, "<<=");
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_sr, ">>=");
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_bw_or, "|=");
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_bw_and, "&=");
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_bw_xor, "^=");
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_pow, "**=");
-#endif
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_inc, "");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_inc, "");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_dec, "");
@@ -642,7 +617,6 @@ XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_concat, ".=");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_dim, "=");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_obj, "=");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_ref, "=&");
-#if PHP_VERSION_ID >= 70400
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_obj_ref, "=&");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_static_prop, "=");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_static_prop_ref, "=&");
@@ -650,7 +624,6 @@ XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_inc_static_prop, "");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_dec_static_prop, "");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_inc_static_prop, "");
 XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_dec_static_prop, "");
-#endif
 
 void xdebug_init_tracing_globals(xdebug_tracing_globals_t *xg)
 {
@@ -664,27 +637,12 @@ void xdebug_tracing_minit(INIT_FUNC_ARGS)
 	xdebug_register_with_opcode_multi_handler(ZEND_ASSIGN, xdebug_assign_handler);
 	xdebug_register_with_opcode_multi_handler(ZEND_QM_ASSIGN, xdebug_qm_assign_handler);
 
-#if PHP_VERSION_ID >= 70400
 	xdebug_set_opcode_handler(ZEND_ASSIGN_OP, xdebug_assign_op_handler);
 	xdebug_set_opcode_handler(ZEND_ASSIGN_DIM_OP, xdebug_assign_dim_op_handler);
 	xdebug_set_opcode_handler(ZEND_ASSIGN_OBJ_OP, xdebug_assign_obj_op_handler);
 	xdebug_set_opcode_handler(ZEND_ASSIGN_STATIC_PROP_OP, xdebug_assign_static_prop_op_handler);
-#else
-	xdebug_set_opcode_handler(ZEND_ASSIGN_ADD, xdebug_assign_add_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_SUB, xdebug_assign_sub_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_MUL, xdebug_assign_mul_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_DIV, xdebug_assign_div_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_MOD, xdebug_assign_mod_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_POW, xdebug_assign_pow_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_SL, xdebug_assign_sl_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_SR, xdebug_assign_sr_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_CONCAT, xdebug_assign_concat_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_BW_OR, xdebug_assign_bw_or_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_BW_AND, xdebug_assign_bw_and_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_BW_XOR, xdebug_assign_bw_xor_handler);
-#endif
-	xdebug_set_opcode_handler(ZEND_ASSIGN_DIM, xdebug_assign_dim_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_OBJ, xdebug_assign_obj_handler);
+	xdebug_register_with_opcode_multi_handler(ZEND_ASSIGN_DIM, xdebug_assign_dim_handler);
+	xdebug_register_with_opcode_multi_handler(ZEND_ASSIGN_OBJ, xdebug_assign_obj_handler);
 	xdebug_set_opcode_handler(ZEND_ASSIGN_REF, xdebug_assign_ref_handler);
 	xdebug_set_opcode_handler(ZEND_PRE_INC, xdebug_pre_inc_handler);
 	xdebug_set_opcode_handler(ZEND_POST_INC, xdebug_post_inc_handler);
@@ -694,21 +652,21 @@ void xdebug_tracing_minit(INIT_FUNC_ARGS)
 	xdebug_set_opcode_handler(ZEND_POST_INC_OBJ, xdebug_post_inc_obj_handler);
 	xdebug_set_opcode_handler(ZEND_PRE_DEC_OBJ, xdebug_pre_dec_obj_handler);
 	xdebug_set_opcode_handler(ZEND_POST_DEC_OBJ, xdebug_post_dec_obj_handler);
-#if PHP_VERSION_ID >= 70400
 	xdebug_set_opcode_handler(ZEND_ASSIGN_OBJ_REF, xdebug_assign_obj_ref_handler);
-	xdebug_set_opcode_handler(ZEND_ASSIGN_STATIC_PROP, xdebug_assign_static_prop_handler);
+	xdebug_register_with_opcode_multi_handler(ZEND_ASSIGN_STATIC_PROP, xdebug_assign_static_prop_handler);
 	xdebug_set_opcode_handler(ZEND_ASSIGN_STATIC_PROP_REF, xdebug_assign_static_prop_ref_handler);
 	xdebug_set_opcode_handler(ZEND_PRE_INC_STATIC_PROP, xdebug_pre_inc_static_prop_handler);
 	xdebug_set_opcode_handler(ZEND_PRE_DEC_STATIC_PROP, xdebug_pre_dec_static_prop_handler);
 	xdebug_set_opcode_handler(ZEND_POST_INC_STATIC_PROP, xdebug_post_inc_static_prop_handler);
 	xdebug_set_opcode_handler(ZEND_POST_DEC_STATIC_PROP, xdebug_post_dec_static_prop_handler);
-#endif
 }
 
 void xdebug_tracing_register_constants(INIT_FUNC_ARGS)
 {
 	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_APPEND", XDEBUG_TRACE_OPTION_APPEND, CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_COMPUTERIZED", XDEBUG_TRACE_OPTION_COMPUTERIZED, CONST_CS | CONST_PERSISTENT);
+	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_FLAMEGRAPH_COST", XDEBUG_TRACE_OPTION_FLAMEGRAPH_COST, CONST_CS | CONST_PERSISTENT);
+	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_FLAMEGRAPH_MEM", XDEBUG_TRACE_OPTION_FLAMEGRAPH_MEM, CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_HTML", XDEBUG_TRACE_OPTION_HTML, CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_NAKED_FILENAME", XDEBUG_TRACE_OPTION_NAKED_FILENAME, CONST_CS | CONST_PERSISTENT);
 }
@@ -717,6 +675,8 @@ void xdebug_tracing_rinit(void)
 {
 	XG_TRACE(trace_handler) = NULL;
 	XG_TRACE(trace_context) = NULL;
+
+	xdebug_disable_opcache_optimizer();
 }
 
 void xdebug_tracing_post_deactivate(void)
@@ -730,25 +690,25 @@ void xdebug_tracing_post_deactivate(void)
 
 void xdebug_tracing_init_if_requested(zend_op_array *op_array)
 {
-	if (xdebug_lib_start_with_request() || xdebug_lib_start_with_trigger(NULL)) {
+	if (xdebug_lib_start_with_request(XDEBUG_MODE_TRACING) || xdebug_lib_start_with_trigger(XDEBUG_MODE_TRACING, NULL)) {
 		/* In case we do an auto-trace we are not interested in the return
 		 * value, but we still have to free it. */
 		xdfree(xdebug_start_trace(NULL, op_array->filename, XINI_TRACE(trace_options)));
 	}
 }
 
-void xdebug_tracing_execute_ex(int function_nr, function_stack_entry *fse)
+void xdebug_tracing_execute_ex(function_stack_entry *fse)
 {
 	if (fse->filtered_tracing || !XG_TRACE(trace_context)) {
 		return;
 	}
 
 	if (XG_TRACE(trace_handler)->function_entry) {
-		XG_TRACE(trace_handler)->function_entry(XG_TRACE(trace_context), fse, function_nr);
+		XG_TRACE(trace_handler)->function_entry(XG_TRACE(trace_context), fse);
 	}
 }
 
-void xdebug_tracing_execute_ex_end(int function_nr, function_stack_entry *fse, zend_execute_data *execute_data)
+void xdebug_tracing_execute_ex_end(function_stack_entry *fse, zend_execute_data *execute_data)
 {
 	zend_op_array *op_array;
 
@@ -757,7 +717,7 @@ void xdebug_tracing_execute_ex_end(int function_nr, function_stack_entry *fse, z
 	}
 
 	if ((XG_TRACE(trace_handler)->function_exit)) {
-		XG_TRACE(trace_handler)->function_exit(XG_TRACE(trace_context), fse, function_nr);
+		XG_TRACE(trace_handler)->function_exit(XG_TRACE(trace_context), fse);
 	}
 
 	/* Store return value in the trace file */
@@ -773,42 +733,42 @@ void xdebug_tracing_execute_ex_end(int function_nr, function_stack_entry *fse, z
 
 	if (op_array->fn_flags & ZEND_ACC_GENERATOR) {
 		if (XG_TRACE(trace_handler)->generator_return_value) {
-			XG_TRACE(trace_handler)->generator_return_value(XG_TRACE(trace_context), fse, function_nr, (zend_generator*) execute_data->return_value);
+			XG_TRACE(trace_handler)->generator_return_value(XG_TRACE(trace_context), fse, (zend_generator*) execute_data->return_value);
 		}
 	} else {
 		if (XG_TRACE(trace_handler)->return_value) {
-			XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, function_nr, execute_data->return_value);
+			XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, execute_data->return_value);
 		}
 	}
 }
 
-int xdebug_tracing_execute_internal(int function_nr, function_stack_entry *fse)
+int xdebug_tracing_execute_internal(function_stack_entry *fse)
 {
 	if (fse->filtered_tracing || !XG_TRACE(trace_context)) {
 		return 0;
 	}
 
 	if (fse->function.type != XFUNC_ZEND_PASS && (XG_TRACE(trace_handler)->function_entry)) {
-		XG_TRACE(trace_handler)->function_entry(XG_TRACE(trace_context), fse, function_nr);
+		XG_TRACE(trace_handler)->function_entry(XG_TRACE(trace_context), fse);
 		return 1;
 	}
 
 	return 0;
 }
 
-void xdebug_tracing_execute_internal_end(int function_nr, function_stack_entry *fse, zval *return_value)
+void xdebug_tracing_execute_internal_end(function_stack_entry *fse, zval *return_value)
 {
 	if (fse->filtered_tracing || !XG_TRACE(trace_context)) {
 		return;
 	}
 
 	if (fse->function.type != XFUNC_ZEND_PASS && (XG_TRACE(trace_handler)->function_exit)) {
-		XG_TRACE(trace_handler)->function_exit(XG_TRACE(trace_context), fse, function_nr);
+		XG_TRACE(trace_handler)->function_exit(XG_TRACE(trace_context), fse);
 	}
 
 	/* Store return value in the trace file */
 	if (XINI_TRACE(collect_return) && fse->function.type != XFUNC_ZEND_PASS && return_value && XG_TRACE(trace_handler)->return_value) {
-		XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, function_nr, return_value);
+		XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, return_value);
 	}
 }
 
