@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2023 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,17 +12,13 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   |          Ilia Alshanetsky <ilia@prohost.org>                         |
-   |          Harald Radi <harald.radi@nme.at>                            |
-   +----------------------------------------------------------------------+
  */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#include "php.h"
+#include "lib/php-header.h"
 #include "main/php_version.h"
 #include "lib/compat.h"
 
@@ -52,11 +48,15 @@
 #include "php_xdebug_arginfo.h"
 
 #include "base/base.h"
+#ifdef __linux__
+# include "base/ctrl_socket.h"
+#endif
 #include "base/filter.h"
 #include "coverage/code_coverage.h"
 #include "develop/monitor.h"
 #include "develop/stack.h"
 #include "develop/superglobals.h"
+#include "debugger/com.h"
 #include "gcstats/gc_stats.h"
 #include "lib/usefulstuff.h"
 #include "lib/lib.h"
@@ -69,17 +69,14 @@
 #include "profiler/profiler.h"
 #include "tracing/tracing.h"
 
-#if PHP_VERSION_ID >= 80000
 static zend_result (*xdebug_orig_post_startup_cb)(void);
 static zend_result xdebug_post_startup(void);
-#elif PHP_VERSION_ID >= 70300
-static int (*xdebug_orig_post_startup_cb)(void);
-static int xdebug_post_startup(void);
-#endif
 
 int xdebug_include_or_eval_handler(zend_execute_data *execute_data);
 
+/* True globals */
 int zend_xdebug_initialised = 0;
+int xdebug_global_mode = 0;
 
 zend_module_entry xdebug_module_entry = {
 	STANDARD_MODULE_HEADER,
@@ -145,26 +142,13 @@ static PHP_INI_MH(OnUpdateSession)
 	DUMP_TOK(session);
 }
 
-static PHP_INI_MH(OnUpdateMode)
-{
-	if (!new_value || !new_value->val) {
-		return FAILURE;
-	}
-
-	if (!xdebug_lib_set_mode(new_value->val)) {
-		return FAILURE;
-	}
-
-	return SUCCESS;
-}
-
 static PHP_INI_MH(OnUpdateStartWithRequest)
 {
-	if (!new_value || !new_value->val) {
+	if (!new_value) {
 		return FAILURE;
 	}
 
-	if (!xdebug_lib_set_start_with_request(new_value->val)) {
+	if (!xdebug_lib_set_start_with_request(ZSTR_VAL(new_value))) {
 		return FAILURE;
 	}
 
@@ -173,11 +157,11 @@ static PHP_INI_MH(OnUpdateStartWithRequest)
 
 static PHP_INI_MH(OnUpdateStartUponError)
 {
-	if (!new_value || !new_value->val) {
+	if (!new_value) {
 		return FAILURE;
 	}
 
-	if (!xdebug_lib_set_start_upon_error(new_value->val)) {
+	if (!xdebug_lib_set_start_upon_error(ZSTR_VAL(new_value))) {
 		return FAILURE;
 	}
 
@@ -189,7 +173,7 @@ static PHP_INI_MH(OnUpdateRemovedSetting)
 	if (! (EG(error_reporting) & E_DEPRECATED)) {
 		return SUCCESS;
 	}
-	if (new_value && new_value->val && strlen(new_value->val) > 0 && strncmp("This setting", new_value->val, 11) != 0) {
+	if (new_value && ZSTR_LEN(new_value) > 0 && strncmp("This setting", ZSTR_VAL(new_value), 11) != 0) {
 		xdebug_log_ex(
 			XLOG_CHAN_CONFIG, XLOG_CRIT, "REMOVED",
 			"The setting '%s' has been removed, see the upgrading guide at %supgrade_guide#changed-%s",
@@ -204,7 +188,7 @@ static PHP_INI_MH(OnUpdateChangedSetting)
 	if (! (EG(error_reporting) & E_DEPRECATED)) {
 		return SUCCESS;
 	}
-	if (new_value && new_value->val && strlen(new_value->val) > 0 && strncmp("This setting", new_value->val, 11) != 0) {
+	if (new_value && ZSTR_LEN(new_value) > 0 && strncmp("This setting", ZSTR_VAL(new_value), 11) != 0) {
 		xdebug_log_ex(
 			XLOG_CHAN_CONFIG, XLOG_CRIT, "CHANGED",
 			"The setting '%s' has been renamed, see the upgrading guide at %supgrade_guide#changed-%s",
@@ -214,6 +198,20 @@ static PHP_INI_MH(OnUpdateChangedSetting)
 	return FAILURE;
 }
 
+#if __linux__
+static PHP_INI_MH(OnUpdateCtrlSocket)
+{
+	if (!new_value) {
+		return FAILURE;
+	}
+
+	if (!xdebug_lib_set_control_socket_granularity(ZSTR_VAL(new_value))) {
+		return FAILURE;
+	}
+
+	return SUCCESS;
+}
+#endif
 
 #ifdef P_tmpdir
 # define XDEBUG_TEMP_DIR P_tmpdir
@@ -225,18 +223,95 @@ static PHP_INI_MH(OnUpdateChangedSetting)
 # endif
 #endif
 
-#define XDEBUG_REMOVED_INI_ENTRY(n) PHP_INI_ENTRY((n), "This setting has been removed, see the upgrading guide at https://xdebug.org/docs/upgrade_guide#removed-" # n, PHP_INI_ALL, OnUpdateRemovedSetting)
-#define XDEBUG_CHANGED_INI_ENTRY(n) PHP_INI_ENTRY((n), "This setting has been changed, see the upgrading guide at https://xdebug.org/docs/upgrade_guide#changed-" # n, PHP_INI_ALL, OnUpdateChangedSetting)
+ZEND_INI_DISP(display_removed_setting)
+{
+	ZEND_PUTS("(setting removed in Xdebug 3)");
+}
+
+ZEND_INI_DISP(display_changed_setting)
+{
+	ZEND_PUTS("(setting renamed in Xdebug 3)");
+}
+
+#if __linux__
+ZEND_INI_DISP(display_control_socket)
+{
+	switch (XINI_BASE(control_socket_granularity))
+	{
+		case XDEBUG_CONTROL_SOCKET_OFF:
+			ZEND_PUTS("off");
+			break;
+		case XDEBUG_CONTROL_SOCKET_TIME:
+			php_printf("time: %ldms", XINI_BASE(control_socket_threshold_ms));
+			break;
+	}
+}
+#endif
+
+#define XDEBUG_REMOVED_INI_ENTRY(n) PHP_INI_ENTRY_EX(("" # n), "This setting has been removed, see the upgrading guide at https://xdebug.org/docs/upgrade_guide#removed-" # n, PHP_INI_ALL, OnUpdateRemovedSetting, display_removed_setting)
+#define XDEBUG_CHANGED_INI_ENTRY(n) PHP_INI_ENTRY_EX(("" # n), "This setting has been changed, see the upgrading guide at https://xdebug.org/docs/upgrade_guide#changed-" # n, PHP_INI_ALL, OnUpdateChangedSetting, display_changed_setting)
+
+static const char *xdebug_start_with_request_types[5] = { "", "default", "yes", "no", "trigger" };
+
+ZEND_INI_DISP(display_start_with_request)
+{
+	char *value;
+
+	if (type == ZEND_INI_DISPLAY_ORIG && ini_entry->modified) {
+		value = ZSTR_VAL(ini_entry->orig_value);
+	} else if (ini_entry->value) {
+		value = ZSTR_VAL(ini_entry->value);
+	} else {
+		value = NULL;
+	}
+	if (value) {
+		ZEND_PUTS(xdebug_start_with_request_types[xdebug_lib_get_start_with_request()]);
+	} else {
+		ZEND_PUTS("?");
+	}
+}
+
+
+static const char *xdebug_start_upon_error_types[4] = { "", "default", "yes", "no" };
+
+ZEND_INI_DISP(display_start_upon_error)
+{
+	char *value;
+
+	if (type == ZEND_INI_DISPLAY_ORIG && ini_entry->modified) {
+		value = ZSTR_VAL(ini_entry->orig_value);
+	} else if (ini_entry->value) {
+		value = ZSTR_VAL(ini_entry->value);
+	} else {
+		value = NULL;
+	}
+	if (value) {
+		ZEND_PUTS(xdebug_start_upon_error_types[xdebug_lib_get_start_upon_error()]);
+	} else {
+		ZEND_PUTS("?");
+	}
+}
+
+
+#if HAVE_XDEBUG_ZLIB
+# define USE_COMPRESSION_DEFAULT "1"
+#else
+# define USE_COMPRESSION_DEFAULT "0"
+#endif
 
 PHP_INI_BEGIN()
 	/* Library settings */
-	PHP_INI_ENTRY(    "xdebug.mode",               "develop",       PHP_INI_SYSTEM,                OnUpdateMode)
-	PHP_INI_ENTRY(    "xdebug.start_with_request", "default",       PHP_INI_SYSTEM|PHP_INI_PERDIR, OnUpdateStartWithRequest)
-	PHP_INI_ENTRY(    "xdebug.start_upon_error",   "default",       PHP_INI_SYSTEM|PHP_INI_PERDIR, OnUpdateStartUponError)
-	STD_PHP_INI_ENTRY("xdebug.output_dir",         XDEBUG_TEMP_DIR, PHP_INI_ALL,                   OnUpdateString, settings.library.output_dir,       zend_xdebug_globals, xdebug_globals)
-	STD_PHP_INI_ENTRY("xdebug.trigger_value",      "",              PHP_INI_SYSTEM|PHP_INI_PERDIR, OnUpdateString, settings.library.trigger_value,    zend_xdebug_globals, xdebug_globals)
-	STD_PHP_INI_ENTRY("xdebug.file_link_format",   "",              PHP_INI_ALL,                   OnUpdateString, settings.library.file_link_format, zend_xdebug_globals, xdebug_globals)
-	STD_PHP_INI_ENTRY("xdebug.filename_format",    "",              PHP_INI_ALL,                   OnUpdateString, settings.library.filename_format,  zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_ENTRY("xdebug.mode",               "develop",               PHP_INI_SYSTEM,                OnUpdateString, settings.library.requested_mode,   zend_xdebug_globals, xdebug_globals)
+	PHP_INI_ENTRY_EX( "xdebug.start_with_request", "default",               PHP_INI_SYSTEM|PHP_INI_PERDIR, OnUpdateStartWithRequest, display_start_with_request)
+	PHP_INI_ENTRY_EX( "xdebug.start_upon_error",   "default",               PHP_INI_SYSTEM|PHP_INI_PERDIR, OnUpdateStartUponError,   display_start_upon_error)
+	STD_PHP_INI_ENTRY("xdebug.output_dir",         XDEBUG_TEMP_DIR,         PHP_INI_ALL,                   OnUpdateString, settings.library.output_dir,       zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_ENTRY("xdebug.use_compression",    USE_COMPRESSION_DEFAULT, PHP_INI_ALL,                   OnUpdateBool,   settings.library.use_compression,  zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_ENTRY("xdebug.trigger_value",      "",                      PHP_INI_SYSTEM|PHP_INI_PERDIR, OnUpdateString, settings.library.trigger_value,    zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_ENTRY("xdebug.file_link_format",   "",                      PHP_INI_ALL,                   OnUpdateString, settings.library.file_link_format, zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_ENTRY("xdebug.filename_format",    "",                      PHP_INI_ALL,                   OnUpdateString, settings.library.filename_format,  zend_xdebug_globals, xdebug_globals)
+#if __linux__
+	PHP_INI_ENTRY_EX("xdebug.control_socket",      "time",                  PHP_INI_ALL,                   OnUpdateCtrlSocket, display_control_socket)
+#endif
 
 	STD_PHP_INI_ENTRY("xdebug.log",       "",           PHP_INI_ALL, OnUpdateString, settings.library.log,       zend_xdebug_globals, xdebug_globals)
 	STD_PHP_INI_ENTRY("xdebug.log_level", XLOG_DEFAULT, PHP_INI_ALL, OnUpdateLong,   settings.library.log_level, zend_xdebug_globals, xdebug_globals)
@@ -247,7 +322,7 @@ PHP_INI_BEGIN()
 	STD_PHP_INI_ENTRY("xdebug.var_display_max_depth",    "3",       PHP_INI_ALL,    OnUpdateLong,   settings.library.display_max_depth,    zend_xdebug_globals, xdebug_globals)
 
 	/* Base settings */
-	STD_PHP_INI_ENTRY("xdebug.max_nesting_level", "256",                PHP_INI_ALL,    OnUpdateLong,   base.settings.max_nesting_level, zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_ENTRY("xdebug.max_nesting_level", "512",                PHP_INI_ALL,    OnUpdateLong,   settings.base.max_nesting_level, zend_xdebug_globals, xdebug_globals)
 
 	/* Develop settings */
 	STD_PHP_INI_ENTRY("xdebug.cli_color",         "0",                  PHP_INI_ALL,    OnUpdateLong,   settings.develop.cli_color,         zend_xdebug_globals, xdebug_globals)
@@ -279,13 +354,13 @@ PHP_INI_BEGIN()
 	/* Xdebug Cloud */
 	STD_PHP_INI_ENTRY("xdebug.cloud_id", "", PHP_INI_SYSTEM, OnUpdateString, settings.debugger.cloud_id, zend_xdebug_globals, xdebug_globals)
 
-	/* Remote debugger settings */
-	STD_PHP_INI_ENTRY("xdebug.client_host",             "localhost", PHP_INI_ALL, OnUpdateString, settings.debugger.client_host,               zend_xdebug_globals, xdebug_globals)
-	STD_PHP_INI_ENTRY("xdebug.client_port",             "9003",      PHP_INI_ALL, OnUpdateLong,   settings.debugger.client_port,               zend_xdebug_globals, xdebug_globals)
-	STD_PHP_INI_BOOLEAN("xdebug.discover_client_host",  "0",         PHP_INI_ALL, OnUpdateBool,   settings.debugger.discover_client_host,      zend_xdebug_globals, xdebug_globals)
-	STD_PHP_INI_ENTRY("xdebug.client_discovery_header", "",          PHP_INI_ALL, OnUpdateString, settings.debugger.client_discovery_header,   zend_xdebug_globals, xdebug_globals)
-	STD_PHP_INI_ENTRY("xdebug.idekey",                  "",          PHP_INI_ALL, OnUpdateString, settings.debugger.ide_key_setting,           zend_xdebug_globals, xdebug_globals)
-	STD_PHP_INI_ENTRY("xdebug.connect_timeout_ms",      "200",       PHP_INI_ALL, OnUpdateLong,   settings.debugger.connect_timeout_ms,        zend_xdebug_globals, xdebug_globals)
+	/* Step debugger settings */
+	STD_PHP_INI_ENTRY("xdebug.client_host",             "localhost",                        PHP_INI_ALL, OnUpdateString, settings.debugger.client_host,             zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_ENTRY("xdebug.client_port",             XDEBUG_CLIENT_PORT_S,               PHP_INI_ALL, OnUpdateLong,   settings.debugger.client_port,             zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_BOOLEAN("xdebug.discover_client_host",  "0",                                PHP_INI_ALL, OnUpdateBool,   settings.debugger.discover_client_host,    zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_ENTRY("xdebug.client_discovery_header", "HTTP_X_FORWARDED_FOR,REMOTE_ADDR", PHP_INI_ALL, OnUpdateString, settings.debugger.client_discovery_header, zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_ENTRY("xdebug.idekey",                  "",                                 PHP_INI_ALL, OnUpdateString, settings.debugger.ide_key_setting,         zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_ENTRY("xdebug.connect_timeout_ms",      "200",                              PHP_INI_ALL, OnUpdateLong,   settings.debugger.connect_timeout_ms,      zend_xdebug_globals, xdebug_globals)
 
 	/* Scream support */
 	STD_PHP_INI_BOOLEAN("xdebug.scream",                 "0",           PHP_INI_ALL,    OnUpdateBool,   settings.develop.do_scream,            zend_xdebug_globals, xdebug_globals)
@@ -298,40 +373,40 @@ PHP_INI_BEGIN()
 	STD_PHP_INI_ENTRY("xdebug.trace_format",      "0",                  PHP_INI_ALL,    OnUpdateLong,   settings.tracing.trace_format,      zend_xdebug_globals, xdebug_globals)
 	STD_PHP_INI_ENTRY("xdebug.trace_options",     "0",                  PHP_INI_ALL,    OnUpdateLong,   settings.tracing.trace_options,     zend_xdebug_globals, xdebug_globals)
 	STD_PHP_INI_BOOLEAN("xdebug.collect_assignments", "0",              PHP_INI_ALL,    OnUpdateBool,   settings.tracing.collect_assignments, zend_xdebug_globals, xdebug_globals)
+	STD_PHP_INI_BOOLEAN("xdebug.collect_params", "1",                   PHP_INI_ALL,    OnUpdateBool,   settings.tracing.collect_params,    zend_xdebug_globals, xdebug_globals)
 	STD_PHP_INI_BOOLEAN("xdebug.collect_return",  "0",                  PHP_INI_ALL,    OnUpdateBool,   settings.tracing.collect_return,    zend_xdebug_globals, xdebug_globals)
 
 	/* Removed/Changed settings */
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.auto_trace")
-	XDEBUG_REMOVED_INI_ENTRY("xdebug.collect_includes")
-	XDEBUG_REMOVED_INI_ENTRY("xdebug.collect_params")
-	XDEBUG_REMOVED_INI_ENTRY("xdebug.collect_vars")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.coverage_enable")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.default_enable")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.gc_stats_enable")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.gc_stats_output_dir")
-	XDEBUG_REMOVED_INI_ENTRY("xdebug.overload_var_dump")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.profiler_enable")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.profiler_enable_trigger")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.profiler_enable_trigger_value")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.profiler_output_dir")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.remote_autostart")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.remote_connect_back")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.remote_enable")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.remote_host")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.remote_log")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.remote_log_level")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.remote_mode")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.remote_port")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.remote_timeout")
-	XDEBUG_REMOVED_INI_ENTRY("xdebug.show_mem_delta")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.trace_output_dir")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.trace_enable_trigger")
-	XDEBUG_CHANGED_INI_ENTRY("xdebug.trace_enable_trigger_value")
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.auto_trace)
+	XDEBUG_REMOVED_INI_ENTRY(xdebug.collect_includes)
+	XDEBUG_REMOVED_INI_ENTRY(xdebug.collect_vars)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.coverage_enable)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.default_enable)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.gc_stats_enable)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.gc_stats_output_dir)
+	XDEBUG_REMOVED_INI_ENTRY(xdebug.overload_var_dump)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.profiler_enable)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.profiler_enable_trigger)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.profiler_enable_trigger_value)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.profiler_output_dir)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.remote_autostart)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.remote_connect_back)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.remote_enable)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.remote_host)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.remote_log)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.remote_log_level)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.remote_mode)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.remote_port)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.remote_timeout)
+	XDEBUG_REMOVED_INI_ENTRY(xdebug.show_mem_delta)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.trace_output_dir)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.trace_enable_trigger)
+	XDEBUG_CHANGED_INI_ENTRY(xdebug.trace_enable_trigger_value)
 PHP_INI_END()
 
-static void xdebug_init_base_globals(struct xdebug_base_info *xg)
+static void xdebug_init_base_globals(xdebug_base_globals_t *xg)
 {
-	xg->level                = 0;
+	xg->stack                = NULL;
 	xg->in_debug_info        = 0;
 	xg->output_is_tty        = OUTPUT_NOT_CHECKED;
 	xg->in_execution         = 0;
@@ -339,17 +414,26 @@ static void xdebug_init_base_globals(struct xdebug_base_info *xg)
 	xg->error_reporting_override   = 0;
 	xg->error_reporting_overridden = 0;
 
-	xg->filter_type_tracing       = XDEBUG_FILTER_NONE;
-	xg->filter_type_profiler      = XDEBUG_FILTER_NONE;
 	xg->filter_type_code_coverage = XDEBUG_FILTER_NONE;
-	xg->filters_tracing           = NULL;
+	xg->filter_type_stack         = XDEBUG_FILTER_NONE;
+	xg->filter_type_tracing       = XDEBUG_FILTER_NONE;
 	xg->filters_code_coverage     = NULL;
+	xg->filters_stack             = NULL;
+	xg->filters_tracing           = NULL;
+
+	xg->php_version_compile_time = PHP_VERSION;
+	xg->php_version_run_time     = zend_get_module_version("standard");
+
+	xdebug_nanotime_init(xg);
 }
 
+
 static void php_xdebug_init_globals(zend_xdebug_globals *xg)
 {
+	memset(&xg->globals, 0, sizeof(xg->globals));
+
 	xdebug_init_library_globals(&xg->globals.library);
-	xdebug_init_base_globals(&xg->base);
+	xdebug_init_base_globals(&xg->globals.base);
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_COVERAGE)) {
 		xdebug_init_coverage_globals(&xg->globals.coverage);
@@ -371,9 +455,11 @@ static void php_xdebug_init_globals(zend_xdebug_globals *xg)
 	}
 }
 
-static void php_xdebug_shutdown_globals (zend_xdebug_globals *xg)
+static void php_xdebug_shutdown_globals(zend_xdebug_globals *xg)
 {
-	xdebug_deinit_develop_globals(&xg->globals.develop);
+	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
+		xdebug_deinit_develop_globals(&xg->globals.develop);
+	}
 }
 
 
@@ -467,6 +553,11 @@ PHP_MINIT_FUNCTION(xdebug)
 	ZEND_INIT_MODULE_GLOBALS(xdebug, php_xdebug_init_globals, php_xdebug_shutdown_globals);
 	REGISTER_INI_ENTRIES();
 
+	/* Locking in mode as it currently is */
+	if (!xdebug_lib_set_mode(XG(settings.library.requested_mode))) {
+		xdebug_lib_set_mode("develop");
+	}
+
 	if (XDEBUG_MODE_IS_OFF()) {
 		return SUCCESS;
 	}
@@ -514,6 +605,9 @@ PHP_MINIT_FUNCTION(xdebug)
 PHP_MSHUTDOWN_FUNCTION(xdebug)
 {
 	if (XDEBUG_MODE_IS_OFF()) {
+#ifdef ZTS
+		ts_free_id(xdebug_globals_id);
+#endif
 		return SUCCESS;
 	}
 
@@ -558,13 +652,6 @@ PHP_RINIT_FUNCTION(xdebug)
 		return SUCCESS;
 	}
 
-#if PHP_VERSION_ID >= 70300 && PHP_VERSION_ID <= 70301
-	/* PHP Bug #77287 causes Xdebug to segfault if Opcache has the "compact
-	 * literals" optimisation turned on. So force the optimisation off for PHP
-	 * 7.3.0 and 7.3.1. */
-	xdebug_disable_opcache_optimizer();
-#endif
-
 	xdebug_library_rinit();
 
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_COVERAGE)) {
@@ -615,9 +702,6 @@ ZEND_MODULE_POST_ZEND_DEACTIVATE_D(xdebug)
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
 		xdebug_develop_post_deactivate();
 	}
-	if (XDEBUG_MODE_IS(XDEBUG_MODE_GCSTATS)) {
-		xdebug_gcstats_post_deactivate();
-	}
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_PROFILING)) {
 		xdebug_profiler_post_deactivate();
 	}
@@ -637,6 +721,13 @@ PHP_RSHUTDOWN_FUNCTION(xdebug)
 		return SUCCESS;
 	}
 
+	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
+		xdebug_develop_rshutdown();
+	}
+	if (XDEBUG_MODE_IS(XDEBUG_MODE_GCSTATS)) {
+		xdebug_gcstats_rshutdown();
+	}
+
 	xdebug_base_rshutdown();
 
 	return SUCCESS;
@@ -674,8 +765,17 @@ ZEND_DLEXPORT void xdebug_statement_call(zend_execute_data *frame)
 
 	lineno = EG(current_execute_data)->opline->lineno;
 
-	xdebug_coverage_count_line_if_active(op_array, op_array->filename, lineno);
-	xdebug_debugger_statement_call(op_array->filename, lineno);
+	if (XDEBUG_MODE_IS(XDEBUG_MODE_COVERAGE)) {
+		xdebug_coverage_count_line_if_active(op_array, op_array->filename, lineno);
+	}
+
+	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
+		xdebug_debugger_statement_call(op_array->filename, lineno);
+	}
+
+#ifdef __linux__
+	xdebug_control_socket_dispatch();
+#endif
 }
 
 ZEND_DLEXPORT int xdebug_zend_startup(zend_extension *extension)
@@ -685,14 +785,13 @@ ZEND_DLEXPORT int xdebug_zend_startup(zend_extension *extension)
 
 	zend_xdebug_initialised = 1;
 
-#if PHP_VERSION_ID >= 70300
 	xdebug_orig_post_startup_cb = zend_post_startup_cb;
 	zend_post_startup_cb = xdebug_post_startup;
 
 	return zend_startup_module(&xdebug_module_entry);
 }
 
-static int xdebug_post_startup(void)
+static zend_result xdebug_post_startup(void)
 {
 	if (xdebug_orig_post_startup_cb) {
 		int (*cb)(void) = xdebug_orig_post_startup_cb;
@@ -706,9 +805,6 @@ static int xdebug_post_startup(void)
 	xdebug_base_post_startup();
 
 	return SUCCESS;
-#else
-	return zend_startup_module(&xdebug_module_entry);
-#endif
 }
 
 ZEND_DLEXPORT void xdebug_zend_shutdown(zend_extension *extension)
